import unittest

from pynusmv.dd import BDD
from pynusmv.init import init_nusmv, deinit_nusmv
from pynusmv.mc import eval_simple_expression

from pynusmv_lower_interface.nusmv.dd import dd as nsdd

from pynusmv_tools.mas import glob

from pynusmv_tools.atlkPO.check import check
from pynusmv_tools.atlkFO.parsing import parseATLK
from pynusmv_tools.atlkPO import config

from pynusmv.model import *
from pynusmv_tools.mas.mas import Agent, Group


class TestCheck(unittest.TestCase):
    
    def setUp(self):
        init_nusmv()
    
    def tearDown(self):
        glob.reset_globals()
        deinit_nusmv()
    
    
    def little(self):
        glob.load_from_file("tests/pynusmv_tools/atlkPO/models/little.smv")
        fsm = glob.mas()
        self.assertIsNotNone(fsm)
        return fsm
        
    def cardgame(self):
        glob.load_from_file("tests/pynusmv_tools/atlkPO/models/cardgame.smv")
        fsm = glob.mas()
        self.assertIsNotNone(fsm)
        return fsm
        
    def cardgame_fair(self):
        glob.load_from_file("tests/pynusmv_tools/atlkPO/models/cardgame-fair.smv")
        fsm = glob.mas()
        self.assertIsNotNone(fsm)
        return fsm
        
    def cardgame_post_fair(self):
        glob.load_from_file("tests/pynusmv_tools/atlkPO/models/cardgame-post-fair.smv")
        fsm = glob.mas()
        self.assertIsNotNone(fsm)
        return fsm
        
    def trans2_fair(self):
        glob.load_from_file("tests/pynusmv_tools/atlkPO/models/"
                            "2-transmission-fair.smv")
        fsm = glob.mas()
        self.assertIsNotNone(fsm)
        return fsm
        
    def transmission(self):
        glob.load_from_file("tests/pynusmv_tools/atlkPO/models/transmission.smv")
        fsm = glob.mas()
        self.assertIsNotNone(fsm)
        return fsm
        
    def transmission_with_knowledge(self):
        glob.load_from_file(
                        "tests/pynusmv_tools/atlkPO/models/transmission-knowledge.smv")
        fsm = glob.mas()
        self.assertIsNotNone(fsm)
        return fsm
        
    def transmission_fair(self):
        glob.load_from_file("tests/pynusmv_tools/atlkPO/models/transmission-fair.smv")
        fsm = glob.mas()
        self.assertIsNotNone(fsm)
        return fsm
    
    def transmission_post_fair(self):
        glob.load_from_file("tests/pynusmv_tools/atlkPO/models/"
                            "transmission-post-fair.smv")
        fsm = glob.mas()
        self.assertIsNotNone(fsm)
        return fsm
    
    def coins(self):
        # Build model
        class main(Module):
            c1 = Var(Scalar(('none', 'head', 'tail')))
            c2 = Var(Scalar(('none', 'head', 'tail')))
            result = Var(Scalar(('none', 'win', 'lose')))

            act1 = IVar(Scalar(("different", "equal")))
            act2 = IVar(Scalar(("different", "equal")))

            INIT = [(c1 == 'none') & (c2 == 'none') & (result == 'none')]

            TRANS = []

            TRANS += [((c1 == 'none') & (c2 == 'none')).implies(
                       c1.next().in_(Set(('head', 'tail'))) &
                       c2.next().in_(Set(('head', 'tail'))) &
                       (result.next() == result))]
            TRANS += [((c1 != 'none') & (c2 != 'none') &
                       (result == 'none')).implies(
                      (c1.next() == c1) & (c2.next() == c2) &
                      (result.next() ==
                       (((c1 == c2) & (act1 == 'equal') & (act2 == 'equal')) |
                        ((c1 != c2) & (act1 == 'different') &
                         (act2 == 'different'))
                       ).ite('win', 'lose')))]
            TRANS += [(result != 'none').implies((result.next() == result) &
                       (c1.next( ) == c1) & (c2.next() == c2))]

        # Define agents
        a1 = Agent("a1", {main.c1}, {main.act1})
        a2 = Agent("a2", {main.c2}, {main.act2})
        g = Group("g", a1, a2)

        # Load model
        glob.load(main)

        # Get MAS
        mas = glob.mas(agents={a1, a2, g})
        
        self.assertIsNotNone(mas)
        return mas
    
    def counters(self):
        # Build model
        class main(Module):
            v1 = Var(Boolean())
            v2 = Var(Boolean())

            act1 = IVar(Scalar(("skip", "inc")))
            act2 = IVar(Scalar(("skip", "inc")))

            TRANS = [v1.next() == (act1 == "skip").ite(v1, ~v1)]
            TRANS += [v2.next() == (act2 == "skip").ite(v2, ~v2)]

        # Define agents
        a1 = Agent("a1", {main.v1}, {main.act1})
        a2 = Agent("a2", {main.v2}, {main.act2})
        g = Group("g", a1, a2)

        # Load model
        glob.load(main)

        # Get MAS
        mas = glob.mas({a1, a2, g})
        
        self.assertIsNotNone(mas)
        return mas
        
    def show_si(self, fsm, bdd):
        if bdd.isnot_false():
            sis = fsm.pick_all_states_inputs(bdd)
            print("SI count:", len(sis))
            for si in sis:
                print(si.get_str_values())
            
    def show_s(self, fsm, bdd):
        if bdd.isnot_false():
            for s in fsm.pick_all_states(bdd):
                print(s.get_str_values())
            
    def show_i(self, fsm, bdd):
        if bdd.isnot_false():
            for i in fsm.pick_all_inputs(bdd):
                print(i.get_str_values())
        
        
    def test_cardgame_not_improved(self):
        fsm = self.cardgame()
        
        self.assertTrue(check(fsm, parseATLK("K<'player'>'pcard=none' & K<'player'>'dcard=none'")[0], implem="partialSI", semantics="individual"))
        self.assertTrue(check(fsm, parseATLK("AG('step = 1' -> ~(K<'player'> 'dcard=Ac' | K<'player'> 'dcard=K' | K<'player'> 'dcard=Q'))")[0], implem="partialSI", semantics="individual"))
        
        self.assertTrue(check(fsm, parseATLK("<'dealer'> X 'pcard=Ac'")[0], implem="partialSI", semantics="individual"))
        self.assertFalse(check(fsm, parseATLK("<'dealer'> G ~'win'")[0], implem="partialSI", semantics="individual"))
        self.assertTrue(check(fsm, parseATLK("['player'] X 'pcard=Ac'")[0], implem="partialSI", semantics="individual"))
        self.assertTrue(check(fsm, parseATLK("['dealer'] F 'win'")[0], implem="partialSI", semantics="individual"))
        self.assertTrue(check(fsm, parseATLK("AG('step = 1' -> ~<'player'> X 'win')")[0], implem="partialSI", semantics="individual"))
        self.assertFalse(check(fsm, parseATLK("<'player'> F 'win'")[0], implem="partialSI", semantics="individual"))
        
        self.assertTrue(check(fsm, parseATLK("EG ~'win'")[0], implem="partialSI", semantics="individual"))
        self.assertTrue(check(fsm, parseATLK("EF 'win'")[0], implem="partialSI", semantics="individual"))
        self.assertFalse(check(fsm, parseATLK("AF 'win'")[0], implem="partialSI", semantics="individual"))
        
    
    def test_cardgame_post_fair_not_improved(self):
        fsm = self.cardgame_post_fair()
        
        self.assertTrue(check(fsm, parseATLK("K<'player'>'pcard=none' & K<'player'>'dcard=none'")[0], implem="partialSI", semantics="individual"))
        self.assertTrue(check(fsm, parseATLK("AG('step = 1' -> ~(K<'player'> 'dcard=Ac' | K<'player'> 'dcard=K' | K<'player'> 'dcard=Q'))")[0], implem="partialSI", semantics="individual"))
        
        self.assertTrue(check(fsm, parseATLK("AG('step = 1' -> ~<'player'> X 'win')")[0], implem="partialSI", semantics="individual"))
        self.assertFalse(check(fsm, parseATLK("['dealer'] F 'win'")[0], implem="partialSI", semantics="individual"))
        self.assertTrue(check(fsm, parseATLK("['player'] X 'pcard=Ac'")[0], implem="partialSI", semantics="individual"))
        self.assertTrue(check(fsm, parseATLK("<'dealer'> X 'pcard=Ac'")[0], implem="partialSI", semantics="individual"))
        self.assertTrue(check(fsm, parseATLK("<'dealer'> G ~'win'")[0], implem="partialSI", semantics="individual"))
        
        # Player can win
        self.assertTrue(check(fsm, parseATLK("<'player'> F 'win'")[0], implem="partialSI", semantics="individual"))
        # Dealer can avoid fairness
        self.assertTrue(check(fsm, parseATLK("<'dealer'> F 'FALSE'")[0], implem="partialSI", semantics="individual"))
        

    def test_transmission_not_improved(self):
        fsm = self.transmission()
        
        self.assertFalse(check(fsm, parseATLK("<'sender'> F 'received'")[0], implem="partialSI", semantics="individual"))
        # False because transmitter cannot win if received is already true
        # and he has no clue about it
        self.assertFalse(check(fsm, parseATLK("<'transmitter'> G ~'received'")[0], implem="partialSI", semantics="individual"))
        self.assertFalse(check(fsm, parseATLK("<'sender'> X 'received'")[0], implem="partialSI", semantics="individual"))
        # False because transmitter cannot win if received is already true
        # and he has no clue about it
        self.assertFalse(check(fsm, parseATLK("<'transmitter'> X ~'received'")[0], implem="partialSI", semantics="individual"))
        self.assertFalse(check(fsm, parseATLK("<'transmitter'> F 'received'")[0], implem="partialSI", semantics="individual"))
        
        # False because the sender does not know if the bit is already
        # transmitted (and in this case, no strategy can avoid 'received')
        self.assertFalse(check(fsm, parseATLK("<'sender'> G ~'received'")[0], implem="partialSI", semantics="individual"))
        
    
    def test_transmission_with_know_not_improved(self):
        fsm = self.transmission_with_knowledge()
        
        self.assertFalse(check(fsm, parseATLK("<'sender'> F 'received'")[0], implem="partialSI", semantics="individual"))
        self.assertTrue(check(fsm, parseATLK("<'transmitter'> G ~'received'")[0], implem="partialSI", semantics="individual"))
        self.assertFalse(check(fsm, parseATLK("<'sender'> X 'received'")[0], implem="partialSI", semantics="individual"))
        self.assertTrue(check(fsm, parseATLK("<'transmitter'> X ~'received'")[0], implem="partialSI", semantics="individual"))
        self.assertFalse(check(fsm, parseATLK("<'transmitter'> F 'received'")[0], implem="partialSI", semantics="individual"))
        self.assertTrue(check(fsm, parseATLK("<'sender'> G ~'received'")[0], implem="partialSI", semantics="individual"))
    
    
    def test_transmission_post_fair_not_improved(self):
        fsm = self.transmission_post_fair()
        
        self.assertTrue(check(fsm, parseATLK("<'sender'> F 'received'")[0], implem="partialSI", semantics="individual"))
        self.assertTrue(check(fsm, parseATLK("<'transmitter'> G ~'received'")[0], implem="partialSI", semantics="individual"))
        self.assertFalse(check(fsm, parseATLK("<'sender'> X 'received'")[0], implem="partialSI", semantics="individual"))
        self.assertTrue(check(fsm, parseATLK("<'transmitter'> X ~'received'")[0], implem="partialSI", semantics="individual"))
        self.assertTrue(check(fsm, parseATLK("<'transmitter'> F 'received'")[0], implem="partialSI", semantics="individual"))
        
        # False because the sender does not know if the bit is already
        # transmitted (and in this case, no strategy can avoid 'received')
        self.assertFalse(check(fsm, parseATLK("<'sender'> G ~'received'")[0], implem="partialSI", semantics="individual"))
        
        
    def test_transmission_improved(self):
        fsm = self.transmission()
        
        # False because transmitter cannot win if received is already true
        # and he has no clue about it
        self.assertFalse(check(fsm, parseATLK("<'transmitter'> G ~'received'")[0], variant="FS", implem="partialSI", semantics="individual"))
        self.assertFalse(check(fsm, parseATLK("<'sender'> X 'received'")[0], variant="FS", implem="partialSI", semantics="individual"))
        # False because transmitter cannot win if received is already true
        # and he has no clue about it
        self.assertFalse(check(fsm, parseATLK("<'transmitter'> X ~'received'")[0], variant="FS", implem="partialSI", semantics="individual"))
        
        # False because the sender does not know if the bit is already
        # transmitted (and in this case, no strategy can avoid 'received')
        self.assertFalse(check(fsm, parseATLK("<'sender'> G ~'received'")[0], variant="FS", implem="partialSI", semantics="individual"))
        
        self.assertFalse(check(fsm, parseATLK("<'transmitter'> F 'received'")[0], variant="FS", implem="partialSI", semantics="individual"))
        self.assertFalse(check(fsm, parseATLK("<'sender'> F 'received'")[0], variant="FS", implem="partialSI", semantics="individual"))
        
    
    def test_transmission_with_know_improved(self):
        fsm = self.transmission_with_knowledge()
        
        self.assertFalse(check(fsm, parseATLK("<'sender'> F 'received'")[0], variant="FS", implem="partialSI", semantics="individual"))
        self.assertTrue(check(fsm, parseATLK("<'transmitter'> G ~'received'")[0], variant="FS", implem="partialSI", semantics="individual"))
        self.assertFalse(check(fsm, parseATLK("<'sender'> X 'received'")[0], variant="FS", implem="partialSI", semantics="individual"))
        self.assertTrue(check(fsm, parseATLK("<'transmitter'> X ~'received'")[0], variant="FS", implem="partialSI", semantics="individual"))
        self.assertFalse(check(fsm, parseATLK("<'transmitter'> F 'received'")[0], variant="FS", implem="partialSI", semantics="individual"))
        self.assertTrue(check(fsm, parseATLK("<'sender'> G ~'received'")[0], variant="FS", implem="partialSI", semantics="individual"))
        
    
    def test_transmission_post_fair_improved(self):
        fsm = self.transmission_post_fair()
        
        self.assertTrue(check(fsm, parseATLK("<'sender'> F 'received'")[0], variant="FS", implem="partialSI", semantics="individual"))
        self.assertTrue(check(fsm, parseATLK("<'transmitter'> G ~'received'")[0], variant="FS", implem="partialSI", semantics="individual"))
        self.assertFalse(check(fsm, parseATLK("<'sender'> X 'received'")[0], variant="FS", implem="partialSI", semantics="individual"))
        self.assertTrue(check(fsm, parseATLK("<'transmitter'> X ~'received'")[0], variant="FS", implem="partialSI", semantics="individual"))
        self.assertTrue(check(fsm, parseATLK("<'transmitter'> F 'received'")[0], variant="FS", implem="partialSI", semantics="individual"))
        
        # False because the sender does not know if the bit is already
        # transmitted (and in this case, no strategy can avoid 'received')
        self.assertFalse(check(fsm, parseATLK("<'sender'> G ~'received'")[0], variant="FS", implem="partialSI", semantics="individual"))
        
        
    def test_cardgame_improved(self):
        fsm = self.cardgame()
        
        self.assertTrue(check(fsm, parseATLK("<'dealer'> X 'pcard=Ac'")[0], variant="FS", implem="partialSI", semantics="individual"))
        self.assertFalse(check(fsm, parseATLK("<'dealer'> G ~'win'")[0], variant="FS", implem="partialSI", semantics="individual"))
        self.assertTrue(check(fsm, parseATLK("AG('step = 1' -> ~<'player'> X 'win')")[0], variant="FS", implem="partialSI", semantics="individual"))
        self.assertFalse(check(fsm, parseATLK("<'player'> F 'win'")[0], variant="FS", implem="partialSI", semantics="individual"))
        
        self.assertTrue(check(fsm, parseATLK("['player'] X 'pcard=Ac'")[0], variant="FS", implem="partialSI", semantics="individual"))
        self.assertTrue(check(fsm, parseATLK("['dealer'] F 'win'")[0], variant="FS", implem="partialSI", semantics="individual"))
        
        
    def test_cardgame_post_fair_improved(self):
        fsm = self.cardgame_post_fair()
        
        self.assertTrue(check(fsm, parseATLK("K<'player'>'pcard=none' & K<'player'>'dcard=none'")[0], variant="FS", implem="partialSI", semantics="individual"))
        self.assertTrue(check(fsm, parseATLK("AG('step = 1' -> ~(K<'player'> 'dcard=Ac' | K<'player'> 'dcard=K' | K<'player'> 'dcard=Q'))")[0], variant="FS", implem="partialSI", semantics="individual"))
        
        self.assertTrue(check(fsm, parseATLK("AG('step = 1' -> ~<'player'> X 'win')")[0], variant="FS", implem="partialSI", semantics="individual"))
        self.assertFalse(check(fsm, parseATLK("['dealer'] F 'win'")[0], variant="FS", implem="partialSI", semantics="individual"))
        self.assertTrue(check(fsm, parseATLK("['player'] X 'pcard=Ac'")[0], variant="FS", implem="partialSI", semantics="individual"))
        self.assertTrue(check(fsm, parseATLK("<'dealer'> X 'pcard=Ac'")[0], variant="FS", implem="partialSI", semantics="individual"))
        self.assertTrue(check(fsm, parseATLK("<'dealer'> G ~'win'")[0], variant="FS", implem="partialSI", semantics="individual"))
        
        # Player can win
        self.assertTrue(check(fsm, parseATLK("<'player'> F 'win'")[0], variant="FS", implem="partialSI", semantics="individual"))
        # Dealer can avoid fairness
        self.assertTrue(check(fsm, parseATLK("<'dealer'> F 'FALSE'")[0], variant="FS", implem="partialSI", semantics="individual"))
    
    def test_coins(self):
        fsm = self.coins()
        
        from pynusmv_tools.atlkPO.evalGen import split
        agents = {'a1', 'a2'}
        strats = split(fsm, fsm.protocol(agents) & fsm.reachable_states,
                       agents, semantics='individual')
        self.assertEqual(len(list(strats)), 64)
        
        self.assertFalse(check(fsm, parseATLK("<'a1', 'a2'> F 'result = win'")[0], implem="partialSI", semantics="individual"))
        self.assertFalse(check(fsm, parseATLK("<'a1', 'a2'> F 'result = win'")[0], implem="partialSI", semantics="individual", variant="FS"))
        
        config.partial.filtering = True
        self.assertFalse(check(fsm, parseATLK("<'a1', 'a2'> F 'result = win'")[0], implem="partialSI", semantics="individual"))
        self.assertFalse(check(fsm, parseATLK("<'a1', 'a2'> F 'result = win'")[0], implem="partialSI", semantics="individual", variant="FS"))
        config.partial.filtering = False
    
    def test_counters(self):
        fsm = self.counters()
        
        from pynusmv_tools.atlkPO.evalGen import split
        agents = {'a1', 'a2'}
        strats = split(fsm, fsm.protocol(agents) & fsm.reachable_states,
                       agents, semantics='individual')
        self.assertEqual(len(list(strats)), 16)
        
        self.assertTrue(check(fsm, parseATLK("<'a1', 'a2'> F 'v1 & v2'")[0], semantics="individual", implem="partialSI"))
        self.assertFalse(check(fsm, parseATLK("<'a1', 'a2'> G '!v1 & !v2'")[0], semantics="individual", implem="partialSI"))
        self.assertTrue(check(fsm, parseATLK("AG('!v1 & !v2' -> <'a1', 'a2'> G '!v1 | !v2')")[0], semantics="individual", implem="partialSI"))
    
    def test_counters_FS(self):
        fsm = self.counters()
        
        from pynusmv_tools.atlkPO.evalGen import split
        agents = {'a1', 'a2'}
        strats = split(fsm, fsm.protocol(agents) & fsm.reachable_states,
                       agents, semantics='individual')
        self.assertEqual(len(list(strats)), 16)
        
        self.assertTrue(check(fsm, parseATLK("<'a1', 'a2'> F 'v1 & v2'")[0], variant="FS", semantics="individual", implem="partialSI"))
        self.assertFalse(check(fsm, parseATLK("<'a1', 'a2'> G '!v1 & !v2'")[0], variant="FS", semantics="individual", implem="partialSI"))
        self.assertTrue(check(fsm, parseATLK("AG('!v1 & !v2' -> <'a1', 'a2'> G '!v1 | !v2')")[0], variant="FS", semantics="individual", implem="partialSI"))
    
    def test_counters_FSF(self):
        fsm = self.counters()
        
        from pynusmv_tools.atlkPO.evalGen import split
        agents = {'a1', 'a2'}
        strats = split(fsm, fsm.protocol(agents) & fsm.reachable_states,
                       agents, semantics='individual')
        self.assertEqual(len(list(strats)), 16)
        
        config.partial.filtering = True
        self.assertTrue(check(fsm, parseATLK("<'a1', 'a2'> F 'v1 & v2'")[0], semantics="individual", implem="partialSI"))
        self.assertFalse(check(fsm, parseATLK("<'a1', 'a2'> G '!v1 & !v2'")[0], semantics="individual", implem="partialSI"))
        self.assertTrue(check(fsm, parseATLK("AG('!v1 & !v2' -> <'a1', 'a2'> G '!v1 | !v2')")[0], semantics="individual", implem="partialSI"))
        config.partial.filtering = False


import unittest

from pynusmv.dd import BDD
from pynusmv.init import init_nusmv, deinit_nusmv
from pynusmv.mc import eval_simple_expression

from pynusmv_tools.mas import glob

from pynusmv_tools.atlkFO.check import check
from pynusmv_tools.atlkFO.parsing import parseATLK


class TestCheck(unittest.TestCase):
    
    def setUp(self):
        init_nusmv()
    
    def tearDown(self):
        glob.reset_globals()
        deinit_nusmv()
        
    def cardgame(self):
        glob.load_from_file("tests/pynusmv_tools/atlkFO/models/cardgame.smv")
        fsm = glob.mas()
        self.assertIsNotNone(fsm)
        return fsm
        
    def transmission(self):
        glob.load_from_file("tests/pynusmv_tools/atlkFO/models/transmission.smv")
        fsm = glob.mas()
        self.assertIsNotNone(fsm)
        return fsm
    
    def tictactoe(self):
        glob.load_from_file("tests/pynusmv_tools/atlkFO/models/tictactoe.smv")
        fsm = glob.mas()
        self.assertIsNotNone(fsm)
        return fsm
        
    def show_si(self, fsm, bdd):
        if bdd.isnot_false():
            sis = fsm.pick_all_states_inputs(bdd)
            print("SI count:", len(sis))
            for si in sis:
                print(si.get_str_values())
            
    def show_s(self, fsm, bdd):
        if bdd.isnot_false():
            for s in fsm.pick_all_states(bdd):
                print(s.get_str_values())
            
    def show_i(self, fsm, bdd):
        if bdd.isnot_false():
            for i in fsm.pick_all_inputs(bdd):
                print(i.get_str_values())
        
    

    def test_cardgame(self):
        fsm = self.cardgame()
        
        self.assertTrue(check(fsm, parseATLK("<'dealer'> X 'pcard=Ac'")[0]))
        self.assertFalse(check(fsm, parseATLK("<'dealer'> G ~'win'")[0]))
        self.assertTrue(check(fsm, parseATLK("['player'] X 'pcard=Ac'")[0]))
        self.assertTrue(check(fsm, parseATLK("['dealer'] F 'win'")[0]))
        self.assertTrue(check(fsm, parseATLK("<'player'> F 'win'")[0]))
        

    def test_transmission(self):
        fsm = self.transmission()
        
        self.assertFalse(check(fsm, parseATLK("<'sender'> F 'received'")[0]))
        self.assertTrue(check(fsm, parseATLK(
                                           "<'transmitter'> G ~'received'")[0]))
        self.assertFalse(check(fsm, parseATLK("<'sender'> X 'received'")[0]))
        self.assertTrue(check(fsm, parseATLK(
                                           "<'transmitter'> X ~'received'")[0]))
        self.assertFalse(check(fsm, parseATLK(
                                            "<'transmitter'> F 'received'")[0]))
        self.assertTrue(check(fsm, parseATLK("<'sender'> G ~'received'")[0]))

                             
                             

    def test_tictactoe(self):
        fsm = self.tictactoe()
        
        self.assertTrue(check(fsm, parseATLK(
                     "'run = circle' -> <'circlep'> X 'board[1] = circle'")[0]))
        self.assertTrue(check(fsm, parseATLK(
                           "'run = circle' -> ['circlep'] X 'run = cross'")[0]))
        self.assertTrue(check(fsm, parseATLK(
                             "<'circlep'> F 'run = stop'")[0]))
        self.assertTrue(check(fsm, parseATLK(
                        "<'circlep', 'crossp'> G 'winner = empty'")[0]))
        self.assertTrue(check(fsm, parseATLK(
                        "<'circlep'> G 'winner != circle'")[0]))


"""
ATLK with partial observability evaluation functions.

These evaluation functions reimplement the algorithms with generators to
decrease the memory consumption of the implementation.
"""

import gc
from functools import reduce

from pynusmv.dd import BDD
from pynusmv.mc import eval_simple_expression
from pynusmv.utils import fixpoint as fp

from ..mas.mas import Agent, Group

from ..atlkFO.ast import (TrueExp, FalseExp, Init, Reachable,
                          Atom, Not, And, Or, Implies, Iff, 
                          AF, AG, AX, AU, AW, EF, EG, EX, EU, EW,
                          nK, nE, nD, nC, K, E, D, C,
                          CEF, CEG, CEX, CEU, CEW, CAF, CAG, CAX, CAU, CAW)
                          

from ..atlkFO.eval import (fair_states, ex, eg, eu, nk, ne, nd, nc)

from . import config


# A dictionary to keep track of number of strategies for improved algorithm
__strategies = {} 
# A dictionary to keep track of number of filterings for improved algorithm
__filterings = {} 


def evalATLK(fsm, spec, variant="SF", semantics="group"):
    """
    Return the BDD representing the set of states of fsm satisfying spec.
    
    fsm -- a MAS representing the system
    spec -- an AST-based ATLK specification
    variant -- the variant of the algorithm to evaluate strategic operators;
               must be
               * "SF" for the standard way: splitting in uniform strategies then
                 filtering winning states,
               * "FS" for the alternating way: filtering winning states, then
                 splitting one conflicting equivalence class, then recurse
               * "FSF" for the filter-split-filter way: filtering winning states
                 then splitting all remaining actions into uniform strategies,
                 then filtering final winning states.
    semantics -- the semantics to use for starting point and strategy point
                 equivalence; must be
                 * "group" for the original ATLK_irF semantics considering
                   the group as a single agent (distributed knowledge is used)
                 * "individual" for the original ATL_ir semantics considering
                   the group as individual agents (individual knowledge is
                   used)
    
    If variant is not in {"SF", "FS", "FSF"}, the standard "SF" way is used.
    If semantics is not in {"group", "individual"}, the "group" semantics is
    used.
    """
    
    if type(spec) is TrueExp:
        return BDD.true(fsm.bddEnc.DDmanager)
        
    elif type(spec) is FalseExp:
        return BDD.false(fsm.bddEnc.DDmanager)
        
    elif type(spec) is Init:
        return fsm.init
        
    elif type(spec) is Reachable:
        return fsm.reachable_states
    
    elif type(spec) is Atom:
        return eval_simple_expression(fsm, spec.value)
        
    elif type(spec) is Not:
        return ~evalATLK(fsm, spec.child, variant=variant, semantics=semantics)
        
    elif type(spec) is And:
        return (evalATLK(fsm, spec.left, variant=variant, semantics=semantics)
                & evalATLK(fsm, spec.right, variant=variant,
                           semantics=semantics))
        
    elif type(spec) is Or:
        return (evalATLK(fsm, spec.left, variant=variant, semantics=semantics)
                | evalATLK(fsm, spec.right, variant=variant,
                           semantics=semantics))
        
    elif type(spec) is Implies:
        # a -> b = ~a | b
        return ((~evalATLK(fsm, spec.left, variant=variant,
                           semantics=semantics))
                 | evalATLK(fsm, spec.right, variant=variant,
                            semantics=semantics))
        
    elif type(spec) is Iff:
        # a <-> b = (a & b) | (~a & ~b)
        l = evalATLK(fsm, spec.left, variant=variant, semantics=semantics)
        r = evalATLK(fsm, spec.right, variant=variant, semantics=semantics)
        return (l & r) | ((~l) & (~r))
        
    elif type(spec) is EX:
        return ex(fsm, evalATLK(fsm, spec.child, variant=variant,
                                semantics=semantics))
        
    elif type(spec) is AX:
        # AX p = ~EX ~p
        return ~ex(fsm, ~evalATLK(fsm, spec.child, variant=variant,
                                  semantics=semantics))
        
    elif type(spec) is EG:
        return eg(fsm, evalATLK(fsm, spec.child, variant=variant, 
                                semantics=semantics))
        
    elif type(spec) is AG:
        # AG p = ~EF ~p = ~E[ true U ~p ]
        return ~eu(fsm,
                   BDD.true(fsm.bddEnc.DDmanager),
                   ~evalATLK(fsm, spec.child, variant=variant,
                             semantics=semantics))
        
    elif type(spec) is EU:
        return eu(fsm, evalATLK(fsm, spec.left, variant=variant,
                                semantics=semantics),
                       evalATLK(fsm, spec.right, variant=variant, 
                                semantics=semantics))
        
    elif type(spec) is AU:
        # A[p U q] = ~E[~q W ~p & ~q] = ~(E[~q U ~p & ~q] | EG ~q)
        p = evalATLK(fsm, spec.left, variant=variant, semantics=semantics)
        q = evalATLK(fsm, spec.right, variant=variant, semantics=semantics)
        equpq = eu(fsm, ~q, ~q & ~p)
        egq = eg(fsm, ~q)
        return ~(equpq | egq)
        
    elif type(spec) is EF:
        # EF p = E[ true U p ]
        return eu(fsm,
                  BDD.true(fsm.bddEnc.DDmanager),
                  evalATLK(fsm, spec.child, variant=variant,
                           semantics=semantics))
        
    elif type(spec) is AF:
        # AF p = ~EG ~p
        return ~eg(fsm, ~evalATLK(fsm, spec.child, variant=variant, 
                                  semantics=semantics))
        
    elif type(spec) is EW:
        # E[ p W q ] = E[ p U q ] | EG p
        return (eu(fsm, evalATLK(fsm, spec.left, variant=variant,
                                 semantics=semantics),
                        evalATLK(fsm, spec.right, variant=variant,
                                 semantics=semantics)) |
                eg(fsm, evalATLK(fsm, spec.left, variant=variant,
                                 semantics=semantics)))
        
    elif type(spec) is AW:
        # A[p W q] = ~E[~q U ~p & ~q]
        p = evalATLK(fsm, spec.left, variant=variant, semantics=semantics)
        q = evalATLK(fsm, spec.right, variant=variant, semantics=semantics)
        return ~eu(fsm, ~q, ~p & ~q)
        
    elif type(spec) is nK:
        return nk(fsm, spec.agent.value, evalATLK(fsm, spec.child,
                                                  variant=variant,
                                                  semantics=semantics))
        
    elif type(spec) is K:
        # K<'a'> p = ~nK<'a'> ~p
        return ~nk(fsm, spec.agent.value, ~evalATLK(fsm, spec.child,
                                                    variant=variant,
                                                    semantics=semantics))
        
    elif type(spec) is nE:
        return ne(fsm,
                  [a.value for a in spec.group],
                  evalATLK(fsm, spec.child, variant=variant,
                           semantics=semantics))
        
    elif type(spec) is E:
        # E<g> p = ~nE<g> ~p
        return ~ne(fsm,
                   [a.value for a in spec.group],
                   ~evalATLK(fsm, spec.child, variant=variant,
                             semantics=semantics))
        
    elif type(spec) is nD:
        return nd(fsm,
                  [a.value for a in spec.group],
                  evalATLK(fsm, spec.child, variant=variant,
                           semantics=semantics))
        
    elif type(spec) is D:
        # D<g> p = ~nD<g> ~p
        return ~nd(fsm,
                   [a.value for a in spec.group],
                   ~evalATLK(fsm, spec.child, variant=variant,
                             semantics=semantics))
        
    elif type(spec) is nC:
        return nc(fsm,
                  [a.value for a in spec.group],
                  evalATLK(fsm, spec.child, variant=variant,
                           semantics=semantics))
        
    elif type(spec) is C:
        # C<g> p = ~nC<g> ~p
        return ~nc(fsm,
                   [a.value for a in spec.group],
                   ~evalATLK(fsm, spec.child, variant=variant,
                             semantics=semantics))
    
    elif type(spec) is CAX:
        # [g] X p = ~<g> X ~p
        newspec = CEX(spec.group, Not(spec.child))
        return ~evalATLK(fsm, newspec, variant=variant, semantics=semantics)
        
    elif type(spec) is CAG:
        # [g] G p = ~<g> F ~p
        newspec = CEF(spec.group, Not(spec.child))
        return ~evalATLK(fsm, newspec, variant=variant, semantics=semantics)
        
    elif type(spec) is CAU:
        # [g][p U q] = ~<g>[ ~q W ~p & ~q ]
        newspec = CEW(spec.group,
                      Not(spec.right),
                      And(Not(spec.left), Not(spec.right)))
        return ~evalATLK(fsm, newspec, variant=variant, semantics=semantics)
        
    elif type(spec) is CAF:
        # [g] F p = ~<g> G ~p
        newspec = CEG(spec.group, Not(spec.child))
        return ~evalATLK(fsm, newspec, variant=variant, semantics=semantics)
        
    elif type(spec) is CAW:
        # [g][p W q] = ~<g>[~q U ~p & ~q]
        newspec = CEU(spec.group,
                      Not(spec.right),
                      And(Not(spec.left), Not(spec.right)))
        return ~evalATLK(fsm, newspec, variant=variant, semantics=semantics)
                   
    elif type(spec) in {CEX, CEG, CEU, CEF, CEW}:
        if variant == "SF":
            return eval_strat(fsm, spec, semantics=semantics)
        elif variant == "FS":
            global __strategies, __filterings
            __strategies[spec] = 0
            __filterings[spec] = 0
            sat = eval_strat_improved(fsm, spec, semantics=semantics)
            
            # DEBUG Print number of strategies and filterings up to know
            if config.debug:
                print("Eval_strat_FS: {} strategies, {} filterings"
                      .format(__strategies[spec], __filterings[spec]))
            
            return sat
        elif variant == "FSF":
            return eval_strat_FSF(fsm, spec, semantics=semantics)
        else:
            return eval_strat(fsm, spec, semantics=semantics)
        
    else:
        # TODO Generate error
        print("[ERROR] evalATLK: unrecognized specification type", spec)
        return None



def cex_si(fsm, agents, phi, strat=None):
    """
    Return the set of state/inputs pairs of strat satisfying <agents> X phi
    under full observability in strat.
    If strat is None, strat is considered true.
    
    fsm -- a MAS representing the system
    agents -- a list of agents names
    phi -- a BDD representing the set of states of fsm satisfying phi
    strat -- a BDD representing allowed state/inputs pairs, or None
    """
    if not strat:
        strat = BDD.true(fsm.bddEnc.DDmanager)
    
    phi = phi & fsm.bddEnc.statesInputsMask
    
    return fsm.pre_strat_si(phi | nfair_gamma_si(fsm, agents, strat),
                            agents, strat)
    

def ceu_si(fsm, agents, phi, psi, strat=None):
    """
    Return the set of state/inputs pairs of strat satisfying <agents>[phi U psi]
    under full observability in strat.
    If strat is None, strat is considered true.
    
    fsm -- a MAS representing the system
    agents -- a list of agents names
    phi -- a BDD representing the set of states of fsm satisfying phi
    psi -- a BDD representing the set of states of fsm satisfying psi
    strat -- a BDD representing allowed state/inputs pairs, or None
    
    """
    if not strat:
        strat = BDD.true(fsm.bddEnc.DDmanager)
    
    phi = phi & fsm.bddEnc.statesInputsMask & strat
    psi = psi & fsm.bddEnc.statesInputsMask & strat
    
    if len(fsm.fairness_constraints) == 0:
        return fp(lambda Z : psi | (phi & fsm.pre_strat_si(Z, agents, strat)),
                  BDD.false(fsm.bddEnc.DDmanager))
    else:
        nfair = nfair_gamma_si(fsm, agents, strat)
        def inner(Z):
            res = psi
            for f in fsm.fairness_constraints:
                nf = ~f & fsm.bddEnc.statesMask & strat
                res = res | fsm.pre_strat_si(fp(lambda Y :
                                                 (phi | psi | nfair) &
                                                 (Z | nf) &
                                                 (psi |
                                                  fsm.pre_strat_si(Y, agents,
                                                                   strat)
                                                 ),
                                              BDD.true(fsm.bddEnc.DDmanager)),
                                              agents, strat)
            return (psi | phi | nfair) & res
        return fp(inner, BDD.false(fsm.bddEnc.DDmanager))
    

def cew_si(fsm, agents, phi, psi, strat=None):
    """
    Return the set of state/inputs pairs of strat satisfying <agents>[phi W psi]
    under full observability in strat.
    If strat is None, strat is considered true.
    
    fsm -- a MAS representing the system
    agents -- a list of agents names
    phi -- a BDD representing the set of states of fsm satisfying phi
    psi -- a BDD representing the set of states of fsm satisfying psi
    strat -- a BDD representing allowed state/inputs pairs, or None
    
    """
    if not strat:
        strat = BDD.true(fsm.bddEnc.DDmanager)
    
    phi = phi & fsm.bddEnc.statesInputsMask & strat
    psi = psi & fsm.bddEnc.statesInputsMask & strat
    
    nfair = nfair_gamma_si(fsm, agents, strat)
    
    return fp(lambda Y : (psi | phi | nfair) &
                         (psi | fsm.pre_strat_si(Y, agents, strat)),
              BDD.true(fsm.bddEnc.DDmanager))
    
    
def ceg_si(fsm, agents, phi, strat=None):
    """
    Return the set of state/inputs pairs of strat satisfying <agents> G phi
    under full observability in strat.
    If strat is None, strat is considered true.
    
    fsm -- a MAS representing the system
    agents -- a list of agents names
    phi -- a BDD representing the set of states of fsm satisfying phi
    strat -- a BDD representing allowed state/inputs pairs, or None
    
    """
    if not strat:
        strat = BDD.true(fsm.bddEnc.DDmanager)
    
    phi = phi & fsm.bddEnc.statesInputsMask & strat
    
    nfair = nfair_gamma_si(fsm, agents, strat)
    
    return fp(lambda Y : (phi | nfair) & fsm.pre_strat_si(Y, agents, strat),
              BDD.true(fsm.bddEnc.DDmanager))


def nfair_gamma_si(fsm, agents, strat=None):
    """
    Return the set of state/inputs pairs of strat
    in which agents can avoid a fair path in strat.
    If strat is None, it is considered true.
    
    fsm -- the model
    agents -- a list of agents names
    strat -- a BDD representing allowed state/inputs pairs, or None
    
    """
    if not strat:
        strat = BDD.true(fsm.bddEnc.DDmanager)
    
    if len(fsm.fairness_constraints) == 0:
        return BDD.false(fsm.bddEnc.DDmanager)
    else:
        def inner(Z):
            res = BDD.false(fsm.bddEnc.DDmanager)
            for f in fsm.fairness_constraints:
                nf = ~f & fsm.bddEnc.statesMask & strat
                res = res | fsm.pre_strat_si(fp(lambda Y :
                                                 (Z | nf) &
                                                 fsm.pre_strat_si(Y, agents,
                                                                  strat),
                                             BDD.true(fsm.bddEnc.DDmanager)),
                                             agents, strat)
            return res
        return fp(inner, BDD.false(fsm.bddEnc.DDmanager))


def get_equiv_class(fsm, gamma, state, semantics="group"):
    """
    Return the equivalence class for gamma state belongs to. The equivalence
    class depends on semantics; if semantics is individual, the common
    knowledge based equivalence class for gamma containing state is returned;
    otherwise (assuming semantics is group), the distributed knowledge based
    equivalence fo gamma containing state is returned.
    
    fsm -- the model
    gamma -- a set of agents of fsm
    semantics -- the semantic to use (group or individual);
                 if not "individual", "group" semantics is used
    """
    if semantics == "individual":
        old = BDD.false(fsm.bddEnc.DDmanager)
        eq = state
        while old != eq:
            old = eq
            for agent in gamma:
                eq |= (fsm.equivalent_states(old, {agent}) &
                       fsm.reachable_states)
            
        return eq
    else:
        return fsm.equivalent_states(state, gamma) & fsm.reachable_states


def is_conflicting(fsm, eqclass, gamma, semantics="group"):
    """
    Return whether the given equivalence class for gamma is conflicting or not.
    
    fsm -- the model
    eqclass -- a BDD representing a set of states/inputs pairs
    gamma -- a set of agents of fsm
    semantics -- the semantic to use for splitting (group or individual);
                 if not "individual", "group" semantics is used
    
    If semantics is "individual", eqclass must be a group knowledge based
    equivalence class for gamma; otherwise, semantics is assumed to be "group"
    and eqclass must be a distributed knowledge based equivalence class.
    
    """
    if semantics == "individual":
        # Consider each agent separately
        for agent in gamma:
            nagent_cube = (fsm.bddEnc.inputsCube -
                           fsm.inputs_cube_for_agents({agent}))
            agentclass = eqclass
            # Get one equivalence class for agent at a time
            while agentclass.isnot_false():
                si = fsm.pick_one_state_inputs(agentclass)
                s = si.forsome(fsm.bddEnc.inputsCube)
                eqcl = agentclass & get_equiv_class(fsm, {agent}, s)
                
                # Check if the equivalence class is conflicting for agent
                if ((eqcl - (eqcl & si.forsome(fsm.bddEnc.statesCube |
                                               nagent_cube)))
                     .isnot_false()):
                    return True
                
                agentclass -= eqcl
        
        return False
    else:
        si = fsm.pick_one_state_inputs(eqclass)
        ngamma_cube = fsm.bddEnc.inputsCube - fsm.inputs_cube_for_agents(gamma)
        return (eqclass - (eqclass & si.forsome(fsm.bddEnc.statesCube |
                                                ngamma_cube))).isnot_false()

def split_for_one_agent(fsm, strat, agent):
    """
    Split the given set of moves strat into non-conflicting sub-strategies for
    agent.
    
    fsm -- the model
    strat -- a BDD representing a set of states/inputs pairs
    agent -- an agent of fsm
    
    The strategies are restricted to moves of strat instead of general
    strategies for agent.
    """
    
    if strat.is_false():
        yield strat
    
    else:
        nagent_cube = (fsm.bddEnc.inputsCube -
                       fsm.inputs_cube_for_agents({agent}))
        
        s = fsm.pick_one_state(strat)
        eqclass = get_equiv_class(fsm, {agent}, s) & strat
        # semantics here has no impact since individual meets group semantics
        # with one single agent
        
        strat = strat - eqclass
        
        while eqclass.isnot_false():
            si = fsm.pick_one_state_inputs(eqclass)
            ia = si.forsome(fsm.bddEnc.statesCube | nagent_cube)
            ncss = eqclass & ia
            
            eqclass -= ncss
            
            for sstrat in split_for_one_agent(fsm, strat, agent):
                yield sstrat | ncss


def split_conflicting(fsm, eqclass, gamma, semantics="group"):
    """
    Split the given equivalence class for gamma into non conflicting
    sub-strategies.
    
    fsm -- the model
    eqclass -- a BDD representing a set of states/inputs pairs
    gamma -- a set of agents of fsm
    semantics -- the semantic to use for splitting (group or individual);
                 if not "individual", "group" semantics is used
    
    If semantics is "individual", eqclass must be a group knowledge based
    equivalence class for gamma; otherwise, semantics is assumed to be "group"
    and eqclass must be a distributed knowledge based equivalence class.
    
    Generate all splitted non conflicting sub-strategies.
    
    """
    if semantics == "individual":
        if len(gamma) <= 0:
            yield eqclass
        else:
            agent = next(iter(sorted(gamma)))
            agents = gamma - {agent}
            for sstrat in split_conflicting(fsm, eqclass, agents,
                                            semantics="individual"):
                for strat in split_for_one_agent(fsm, sstrat, agent):
                    yield strat
    else:
        ngamma_cube = fsm.bddEnc.inputsCube - fsm.inputs_cube_for_agents(gamma)
        # Split eqclass into non-conflicting subsets
        while eqclass.isnot_false():
            si = fsm.pick_one_state_inputs(eqclass)
            ncss = eqclass & si.forsome(fsm.bddEnc.statesCube | ngamma_cube)
            eqclass = eqclass - ncss

            yield ncss


def split_one(fsm, strats, gamma, semantics="group"):
    """
    Split one equivalence class of strats and return triples composed of
    the common non-conflicting part already encountered,
    a split and the rest of strats to split.
    
    fsm -- the model
    strats -- a BDD representing a set of states/inputs pairs
    gamma -- a set of agents of fsm
    semantics -- the semantic to use for splitting (group or individual);
                 if not "individual", "group" semantics is used
    
    Return a generator of all triples of common parts, splits
    and rest of strats.
    
    """
    if strats.is_false():
        yield (strats, strats, strats)
        return
        
    else:
        common = BDD.false(fsm.bddEnc.DDmanager)
        
        while strats.isnot_false():
            # Get one equivalence class
            si = fsm.pick_one_state_inputs(strats)
            s = si.forsome(fsm.bddEnc.inputsCube)
            eqs = get_equiv_class(fsm, gamma, s, semantics=semantics)
            eqcl = strats & eqs
            
            # Remove it from strats
            strats = strats - eqcl
            
            # The current equivalence class is conflicting
            if is_conflicting(fsm, eqcl, gamma, semantics=semantics):
                for non_conflicting in split_conflicting(fsm, eqcl, gamma,
                                                         semantics=semantics):
                    yield (common, non_conflicting, strats)
                return
            
            else:
                # Add equivalence class to common
                common = common | eqcl
        
        # strats is false, everything is in common
        yield (common, strats, strats)


def split(fsm, strats, gamma, semantics="group"):
    """
    Split strats into all its non-conflicting greatest subsets.
    
    fsm -- the model
    strats -- a BDD representing a set of states/inputs pairs
    gamma -- a set of agents of fsm
    semantics -- the semantic to use for splitting (group or individual)
    
    Return a generator of all non-conflicting greatest subsets of strats.
    
    """
    if strats.is_false():
        yield strats
    else:
        # semantics should be group
        # Split one equivalence class
        for common, splitted, rest in split_one(fsm, strats, gamma,
                                                semantics=semantics):
            for strat in split(fsm, rest, gamma,
                               semantics=semantics):
                yield (common | strat | splitted)


def filter_strat(fsm, spec, strat=None, variant="SF", semantics="group"):
    """
    Returns the subset SA of strat (or the whole system if strat is None),
    state/action pairs of fsm, such that there is a strategy to satisfy spec
    in fsm.
    
    fsm -- a MAS representing the system;
    spec -- an AST-based ATLK specification with a top strategic operator;
            the operator is CEX, CEG, CEF, CEU or CEW.
    strat -- the subset of the system to consider.
    variant -- the variant of the algorithm to evaluate strategic operators;
               must be
               * "SF" for the standard way: splitting in uniform strategies then
                 filtering winning states,
               * "FS" for the alternating way: filtering winning states, then
                 splitting one conflicting equivalence class, then recurse
               * "FSF" for the filter-split-filter way: filtering winning states
                 then splitting all remaining actions into uniform strategies,
                 then filtering final winning states.
    semantics -- the semantics to use for starting point and strategy point
                 equivalence; must be
                 * "group" for the original ATLK_irF semantics considering
                   the group as a single agent (distributed knowledge is used)
                 * "individual" for the original ATL_ir semantics considering
                   the group as individual agents (individual knowledge is
                   used)
                 
    If variant is not in {"SF", "FS", "FSF"}, the standard "SF" way is used.
    """
    
    sat = BDD.false(fsm.bddEnc.DDmanager)
    agents = {atom.value for atom in spec.group}
    
    # Filtering
    if type(spec) is CEX:
        winning = cex_si(fsm, agents,
                         evalATLK(fsm, spec.child, variant=variant,
                                  semantics=semantics), strat)

    elif type(spec) is CEG:
        winning = ceg_si(fsm, agents, evalATLK(fsm, spec.child,
                                               variant=variant,
                                               semantics=semantics), strat)

    elif type(spec) is CEU:
        winning = ceu_si(fsm, agents, evalATLK(fsm, spec.left,
                                               variant=variant,
                                               semantics=semantics),
                         evalATLK(fsm, spec.right, variant=variant,
                                  semantics=semantics), strat)

    elif type(spec) is CEF:
        # <g> F p = <g>[true U p]
        winning = ceu_si(fsm, agents, BDD.true(fsm.bddEnc.DDmanager),
                         evalATLK(fsm, spec.child, variant=variant,
                                  semantics=semantics), strat)

    elif type(spec) is CEW:
       winning = cew_si(fsm, agents, evalATLK(fsm, spec.left, variant=variant,
                                              semantics=semantics),
                        evalATLK(fsm, spec.right, variant=variant,
                                 semantics=semantics), strat)
    
    
    return winning & fsm.bddEnc.statesInputsMask & fsm.protocol(agents)


def all_equiv_sat(fsm, winning, agents, semantics="group"):
    """
    Return the states s of winning such that all states indistinguishable
    from s by agents are in winning.
    
    fsm -- a MAS representing the system;
    winning -- a BDD representing a set of states of fsm;
    agents -- a set of agents;
    semantics -- the semantics to use (group or individual); if not
                 "individual", "group" semantics is used.
    
    """
    if semantics == "individual":
        equiv_sat = BDD.true(fsm.bddEnc.DDmanager)
        for agent in agents:
            equiv_sat &= all_equiv_sat(fsm, winning, {agent},
                                       semantics="group")
        return equiv_sat
    else:
        # Get states for which all states belong to winning
        # wineq is the set of states for which all equiv states are in winning
        nwinning = ~winning & fsm.bddEnc.statesInputsMask
        return ~(fsm.equivalent_states(nwinning &
                 fsm.reachable_states, frozenset(agents))) & winning


def agents_in_group(fsm, group):
    """
    Return the set of agents in the given group in fsm. If group is not a group
    of fsm, group is returned alone.
    
    This procedure recursively searches for all basic agents in the groups
    possibly composing group.
    
    fsm -- the model;
    group -- a group or agent name of the model.
    """
    
    if group in fsm.groups:
        agents = set()
        for agent in fsm.groups[group]:
            if agent in fsm.groups:
                agents |= agents_in_group(fsm, agent)
            else:
                agents.add(agent)
    else:
        agents = {group}
    return agents


def eval_strat(fsm, spec, semantics="group"):
    """
    Return the BDD representing the set of states of fsm satisfying spec.
    spec is a strategic operator <G> pi.
    
    fsm -- a MAS representing the system;
    spec -- an AST-based ATLK specification with a top strategic operator.
    semantics -- the semantics to use (group or individual)
    
    """
    sat = BDD.false(fsm.bddEnc.DDmanager)
    agents = {atom.value for atom in spec.group}
    
    if semantics == "individual":
        agents = reduce(lambda a, b: a | b,
                        (agents_in_group(fsm, group) for group in agents))
    
    # Restrict protocol to reachable states to avoid splitting useless
    # equivalence classes
    protocol = fsm.protocol(agents) & fsm.reachable_states
    strats = split(fsm, protocol, agents, semantics=semantics)
    nbstrats = 0
    
    if config.debug:
        moves_count = fsm.count_states_inputs(protocol)
        print("Eval strategies (SF): {} move{} in protocol"
              .format(moves_count, "s" if moves_count > 1 else ""))
    
    for strat in strats:
        nbstrats += 1
        winning = (filter_strat(fsm, spec, strat, variant="SF").
                    forsome(fsm.bddEnc.inputsCube))
        sat = sat | all_equiv_sat(fsm, winning, agents, semantics=semantics)
        
        # ----- Garbage collection -------------------------------------
        if (config.garbage.type == "each" or
            (config.garbage.type == "step"
                and nbstrats % config.garbage.step == 0)):
            gc.collect()
        
        if config.debug and nbstrats % 1000 == 0:
            print("Eval strategies (SF): {} strateg{} checked so far"
                  .format(nbstrats, "ies" if nbstrats > 1 else "y"))
    
    # DEBUG Print number of strategies
    if config.debug:
        print("Eval_strat: {} strategies".format(nbstrats))
    
    return sat


def eval_strat_improved(fsm, spec, toSplit=None, toKeep=None,
                        semantics="group"):
    """
    Return the BDD representing the set of states of fsm satisfying spec.
    spec is a strategic operator <G> pi.
    
    Implement the improved algorithm alternating between splitting a conflicting
    equvalence class and filtering winning state/inputs pairs.
    
    fsm -- a MAS representing the system;
    spec -- an AST-based ATLK specification with a top strategic operator.
    toSplit -- the part of strategy to split (not necessarily uniform).
    toKeep -- the other part of strategy to keep (uniform).
    semantics -- the semantics to use (group or individual)
    
    """
    sat = BDD.false(fsm.bddEnc.DDmanager)
    agents = {atom.value for atom in spec.group}
    
    if semantics == "individual":
        agents = reduce(lambda a, b: a | b,
                        (agents_in_group(fsm, group) for group in agents))
    
    gamma = agents
    ngamma_cube = fsm.bddEnc.inputsCube - fsm.inputs_cube_for_agents(gamma)
    
    if not toSplit and not toKeep:
        toSplit = fsm.protocol(agents)
        toKeep = BDD.false(fsm.bddEnc.DDmanager)
        
    
    winning = filter_strat(fsm, spec, toSplit | toKeep, variant="FS",
                           semantics=semantics)
    toSplit = toSplit & winning
    toKeep = toKeep & winning
    global __filterings
    __filterings[spec] += 1
    
    # Get one conflicting equivalence class
    if winning.is_false(): # no state/inputs pairs are winning => return false
        return winning
    
    # Split one conflicting equivalence class
    # TODO Consider semantics
    for common, splitted, rest in split_one(fsm, toSplit, gamma,
                                            semantics=semantics):
        
        if splitted.is_false():
            # No conflicting classes, return states that are winning for all eq
            common = (common | toKeep).forsome(fsm.bddEnc.inputsCube)
            sat = sat | all_equiv_sat(fsm, common, agents, semantics=semantics)
            global __strategies
            __strategies[spec] += 1
            
            # Collect to avoid memory overflow
            if (config.garbage.type == "each" or
                (config.garbage.type == "step"
                 and __strategies[spec] % config.garbage.step == 0)):
                gc.collect()
        
        else:
            sat = sat | eval_strat_improved(fsm, spec, 
                                            toSplit=rest,
                                            toKeep=common | splitted | toKeep,
                                            semantics=semantics)
    
    return sat
    
    
def eval_strat_FSF(fsm, spec, semantics="group"):
    """
    Return the BDD representing the set of states of fsm satisfying spec.
    spec is a strategic operator <G> pi.
    
    Implement a variant of the algorithm that
    filters, splits and then filters.
    
    fsm -- a MAS representing the system;
    spec -- an AST-based ATLK specification with a top strategic operator.
    semantics -- the semantics to use (group or individual).
    
    """
    
    sat = BDD.false(fsm.bddEnc.DDmanager)
    agents = {atom.value for atom in spec.group}
    
    if semantics == "individual":
        agents = reduce(lambda a, b: a | b,
                        (agents_in_group(fsm, group) for group in agents))
    
    protocol = fsm.protocol(agents) & fsm.reachable_states
    
    if config.debug:
        moves_count = fsm.count_states_inputs(protocol)
        print("Eval strategies (FSF): {} move{} in protocol"
              .format(moves_count, "s" if moves_count > 1 else ""))
    
    # First filtering
    winning = filter_strat(fsm, spec, strat=protocol, variant="FSF",
                           semantics=semantics)
    
    if config.debug:
        moves_count = fsm.count_states_inputs(winning)
        print("Eval strategies (FSF): {} move{} after filtering"
              .format(moves_count, "s" if moves_count > 1 else ""))
    
    if winning.is_false(): # no state/inputs pairs are winning => return false
        return winning
    
    # Splitting the strategies
    nbstrats = 0
    for strat in split(fsm, winning, agents, semantics=semantics):
        nbstrats += 1
        
        if config.debug and nbstrats % 1000 == 0:
            print("Eval strategies (FSF): {} strateg{} checked so far"
                  .format(nbstrats, "ies" if nbstrats > 1 else "y"))
        
        # Second filtering
        winning = filter_strat(fsm, spec, strat, variant="FSF",
                               semantics=semantics)
        winning = winning.forsome(fsm.bddEnc.inputsCube)
        sat = sat | all_equiv_sat(fsm, winning, agents, semantics=semantics)
        
        # Collect to avoid memory overflow
        if (config.garbage.type == "each" or
            (config.garbage.type == "step"
             and nbstrats % config.garbage.step == 0)):
            gc.collect()
    
    if config.debug:
        print("Eval strategies (FSF): {} strateg{} checked"
              .format(nbstrats, "ies" if nbstrats > 1 else "y"))
    
    return sat


import unittest

from pynusmv.init import init_nusmv, deinit_nusmv
from pynusmv.mc import eval_simple_expression

from pynusmv_tools.mas import glob
from pynusmv_tools.ctlk.parsing import parseCTLK
from pynusmv_tools.ctlk.simulation.stateChoice import (choose_one_state,
                                               choose_next_state)


class TestStateChoice(unittest.TestCase):
    
    def setUp(self):
        init_nusmv()
        
    def tearDown(self):
        glob.reset_globals()
        deinit_nusmv()
        
    def model(self):
        glob.load_from_file("tests/pynusmv_tools/ctlk/dining-crypto.smv")
        fsm = glob.mas()
        self.assertIsNotNone(fsm)
        return fsm
    
    @unittest.skip # Skip it to keep tests automatic 
    def test_one_state(self):
        fsm = self.model()
        c2p = eval_simple_expression(fsm, "c2.payer")
        state = choose_one_state(fsm, fsm.init)
        
        print()
        if state is None:
            print("No chosen state.")
        else:
            values = state.get_str_values()
            for var in values:
                print(var, "=", values[var])

    
    @unittest.skip # Skip it to keep tests automatic         
    def test_next_state_with_inputs(self):
        fsm = self.model()
        c2p = eval_simple_expression(fsm, "c2.payer")
        (inputs, state) = choose_next_state(fsm, fsm.pick_one_state(fsm.init))
        
        print()
        if state is None:
            print("No chosen state.")
        else:
            if inputs is not None:
                values = inputs.get_str_values()
                for var in values:
                    print(var, "=", values[var])
                print("-" * 40)
            values = state.get_str_values()
            for var in values:
                print(var, "=", values[var])
                
    
    @unittest.skip # Skip it to keep tests automatic
    def test_next_state_without_inputs(self):
        glob.load_from_file("tests/pynusmv/models/modules.smv")
        fsm = glob.mas()
        self.assertIsNotNone(fsm)
        c2p = eval_simple_expression(fsm, "top")
        (inputs, state) = choose_next_state(fsm, fsm.pick_one_state(fsm.init))
        
        print()
        if state is None:
            print("No chosen state.")
        else:
            if inputs is not None:
                values = inputs.get_str_values()
                for var in values:
                    print(var, "=", values[var])
                print("-" * 40)
            values = state.get_str_values()
            for var in values:
                print(var, "=", values[var])


'''
This module validates the behavior of the utility functions defined in the 
module :mod:`pynusmv_tools.bmcLTL.check`.
'''

from unittest             import TestCase
from tests                import utils as tests

from pynusmv_tools.bmcLTL.parsing import parseLTL
from pynusmv_tools.bmcLTL         import check # the tested module 

class TestCheck(TestCase):
    
    def test_GLOBALLY_check_ltl_onepb(self):
        with tests.Configure(self, __file__, "/example.smv"):
            formula      = parseLTL("[](a <=> !b)")
            for k in range(10):
                status,trace = check.check_ltl_onepb(formula, k)
                self.assertEqual("Ok", status)
                self.assertIsNone(trace)
             
            # already violated in the initial state
            formula      = parseLTL("[](a <=> b)")
            status,trace = check.check_ltl_onepb(formula, 0)
            self.assertEqual("Violation", status)
            self.assertIsNotNone(trace)
            self.assertEqual(0, len(trace))
                     
                 
    def test_EVENTUALLY_check_ltl_onepb(self):
        with tests.Configure(self, __file__, "/example.smv"):       
            # proving a violation of this prop necessitates at least two
            # steps: flip --> flop --> flip 
            formula      = parseLTL("<>(a <=> b)")
             
            status,trace = check.check_ltl_onepb(formula, 0)
            self.assertEqual("Ok", status)
            self.assertIsNone(trace)
             
            status,trace = check.check_ltl_onepb(formula, 1)
            self.assertEqual("Ok", status)
            self.assertIsNone(trace)
             
            # a loop is identified
            status,trace = check.check_ltl_onepb(formula, 2)
            self.assertEqual("Violation", status)
            self.assertIsNotNone(trace)
            self.assertEqual(2, len(trace))
             
            # valid in the initial state (hence for any bound)
            formula      = parseLTL("<>(a | b)")
            for k in range(10):
                status,trace = check.check_ltl_onepb(formula, k)
                self.assertEqual("Ok", status)
                self.assertIsNone(trace)
                     
    def test_NEXT_check_ltl_onepb(self):
        with tests.Configure(self, __file__, "/example.smv"):       
            # false in the initial state
            formula      = parseLTL("() a")
             
            # however the violation is not detected when no move is allowed
            status,trace = check.check_ltl_onepb(formula, 0)
            self.assertEqual("Ok", status)
            self.assertIsNone(trace)
             
            status,trace = check.check_ltl_onepb(formula, 1)
            self.assertEqual("Violation", status)
            self.assertIsNotNone(trace)
            self.assertEqual(1, len(trace))
             
            # true in the initial state
            formula      = parseLTL("()() a")
            # not reachable
            status,trace = check.check_ltl_onepb(formula, 0)
            self.assertEqual("Ok", status)
            self.assertIsNone(trace)
            # not quite yet
            status,trace = check.check_ltl_onepb(formula, 1)
            self.assertEqual("Ok", status)
            self.assertIsNone(trace)
            # ok
            status,trace = check.check_ltl_onepb(formula, 2)
            self.assertEqual("Ok", status)
            self.assertIsNone(trace)
            # and even for longer traces
            status,trace = check.check_ltl_onepb(formula, 3)
            self.assertEqual("Ok", status)
            self.assertIsNone(trace)
                 
    def test_WEAKUNTIL_check_ltl_onepb(self):
        with tests.Configure(self, __file__, "/never_b.smv"):
            # entailed by the automaton
            formula      = parseLTL("a W b")
             
            # not reachable
            status,trace = check.check_ltl_onepb(formula, 0)
            self.assertEqual("Ok", status)
            self.assertIsNone(trace)
            # already looping but no counter example
            status,trace = check.check_ltl_onepb(formula, 1)
            self.assertEqual("Ok", status)
            self.assertIsNone(trace)
             
            # true in the initial state
            formula      = parseLTL("b W a")
             
            status,trace = check.check_ltl_onepb(formula, 0)
            self.assertEqual("Ok", status)
            self.assertIsNone(trace)
             
            status,trace = check.check_ltl_onepb(formula, 1)
            self.assertEqual("Ok", status)
            self.assertIsNone(trace)
             
            status,trace = check.check_ltl_onepb(formula, 2)
            self.assertEqual("Ok", status)
            self.assertIsNone(trace)
             
            status,trace = check.check_ltl_onepb(formula, 3)
            self.assertEqual("Ok", status)
            self.assertIsNone(trace)
             
            # violated right away
            formula      = parseLTL("b W !a")
             
            status,trace = check.check_ltl_onepb(formula, 0)
            self.assertEqual("Violation", status)
            self.assertIsNotNone(trace)
            self.assertEqual(0, len(trace))
                 
    def test_UNTIL_check_ltl_onepb(self):
        with tests.Configure(self, __file__, "/never_b.smv"):    
            # entailed by the automaton
            formula      = parseLTL("a U b")
             
            # not reachable
            status,trace = check.check_ltl_onepb(formula, 0)
            self.assertEqual("Ok", status)
            self.assertIsNone(trace)
             
            # this is where U differs from W: at infinity, b must hold
            status,trace = check.check_ltl_onepb(formula, 1)
            self.assertEqual("Violation", status)
            self.assertIsNotNone(trace)
            self.assertEqual(1, len(trace))
             
            # true in the initial state
            formula      = parseLTL("b U a")
             
            status,trace = check.check_ltl_onepb(formula, 0)
            self.assertEqual("Ok", status)
            self.assertIsNone(trace)
             
            status,trace = check.check_ltl_onepb(formula, 1)
            self.assertEqual("Ok", status)
            self.assertIsNone(trace)
             
            status,trace = check.check_ltl_onepb(formula, 2)
            self.assertEqual("Ok", status)
            self.assertIsNone(trace)
             
            status,trace = check.check_ltl_onepb(formula, 3)
            self.assertEqual("Ok", status)
            self.assertIsNone(trace)
             
            # violated right away
            formula      = parseLTL("b U !a")
             
            status,trace = check.check_ltl_onepb(formula, 0)
            self.assertEqual("Violation", status)
            self.assertIsNotNone(trace)
            self.assertEqual(0, len(trace))
                 
    def test_GLOBALLY_check_ltl(self):
        with tests.Configure(self, __file__, "/example.smv"):
            status,l,trace = check.check_ltl(parseLTL("[](a <=> !b)"), 5)
            self.assertEqual('Ok', status)
            self.assertEqual(5, l)
            self.assertIsNone(trace)
             
            status,l,trace = check.check_ltl(parseLTL("[](a & b)"), 5)
            self.assertEqual('Violation', status)
            self.assertEqual(0, l)
            self.assertIsNotNone(trace)
                 
    def test_EVENTUALLY_check_ltl(self):
        with tests.Configure(self, __file__, "/example.smv"):
            status,l,trace = check.check_ltl(parseLTL("<>(a <=> !b)"), 5)
            self.assertEqual('Ok', status)
            self.assertEqual(5, l)
            self.assertIsNone(trace)
             
            status,l,trace = check.check_ltl(parseLTL("<>(a & b)"), 5)
            self.assertEqual('Violation', status)
            self.assertEqual(2, l)
            self.assertIsNotNone(trace)
                 
    def test_NEXT_check_ltl(self):
        with tests.Configure(self, __file__, "/example.smv"):
            status,l,trace = check.check_ltl(parseLTL("()() a"), 5)
            self.assertEqual('Ok', status)
            self.assertEqual(5, l)
            self.assertIsNone(trace)
             
            status,l,trace = check.check_ltl(parseLTL("() a"), 5)
            self.assertEqual('Violation', status)
            self.assertEqual(1, l)
            self.assertIsNotNone(trace)
                 
    def test_WEAKUNTIL_check_ltl(self):
        with tests.Configure(self, __file__, "/never_b.smv"):
            status,l,trace = check.check_ltl(parseLTL("a W b"), 5)
            self.assertEqual('Ok', status)
            self.assertEqual(5, l)
            self.assertIsNone(trace)
             
            status,l,trace = check.check_ltl(parseLTL("b W a"), 5)
            self.assertEqual('Ok', status)
            self.assertEqual(5, l)
            self.assertIsNone(trace)
             
            status,l,trace = check.check_ltl(parseLTL("b W !a"), 5)
            self.assertEqual('Violation', status)
            self.assertEqual(0, l)
            self.assertIsNotNone(trace)
                 
    def test_UNTIL_check_ltl(self):
        with tests.Configure(self, __file__, "/never_b.smv"):
            status,l,trace = check.check_ltl(parseLTL("b U a"), 5)
            self.assertEqual('Ok', status)
            self.assertEqual(5, l)
            self.assertIsNone(trace)
             
            status,l,trace = check.check_ltl(parseLTL("a U b"), 5)
            self.assertEqual('Violation', status)
            self.assertEqual(1, l)
            self.assertIsNotNone(trace)
             
            status,l,trace = check.check_ltl(parseLTL("b W !a"), 5)
            self.assertEqual('Violation', status)
            self.assertEqual(0, l)
            self.assertIsNotNone(trace)
                
    def test_check_ltl_fairness(self):
        """
        This tests validates two features::
            1. the use of a non-variable symbol in the formula
            2. the use of the fairness flag when performing the check.
        """
        with tests.Configure(self, __file__, "/philo.smv"):
            # no one must wait forever:
            # -> this is entailed by fair execution but not in general
            formula = parseLTL("(!<>[](p1.waiting)) & (!<>[](p2.waiting))")
            
            # unfair
            status,_,trace = check.check_ltl(formula, 10, no_fairness=True)
            self.assertEqual("Violation", status)
            self.assertEqual(3, len(trace))
            
            # fair exec only
            status,_,trace = check.check_ltl(formula, 10, no_fairness=False)
            self.assertEqual("Ok", status)
            self.assertIsNone(trace)
            
    def test_check_ltl_invariants(self):
        """
        This tests the use of the invariants flag when performing the check.
        """
        with tests.Configure(self, __file__, "/dummy_with_invar.smv"):
            formula = parseLTL("[] v")
            
            # invariants enforced
            status,_,trace = check.check_ltl(formula, 10, no_invar=True)
            self.assertEqual("Violation", status)
            self.assertEqual(0, len(trace))
            
            # invariants enforced
            status,_,trace = check.check_ltl(formula, 10, no_invar=False)
            self.assertEqual("Ok", status)
            self.assertIsNone(trace)
            
    def test_check_ltl_arithmetics(self):
        """
        This tests the use of the invariants flag when performing the check.
        """
        with tests.Configure(self, __file__, "/numbers.smv"):
            formula = parseLTL("[] a < 7")
            
            status,_,trace = check.check_ltl(formula, 10, no_invar=False)
            self.assertEqual("Ok", status)
            self.assertIsNone(trace)
            


from .tlace import PartialTlacenode, EpistemicBranch, TemporalBranch

from ..ast import (TrueExp, FalseExp, Init, Reachable,
                   Atom, Not, And, Or, Implies, Iff, 
                   AF, AG, AX, AU, AW, EF, EG, EX, EU, EW,
                   nK, nE, nD, nC, K, E, D, C)
from ..eval import evalCTLK
from ..explain import (explain_ex, explain_eg, explain_eu,
                       explain_nk, explain_ne, explain_nd, explain_nc,
                       explain_reachable)


"""
Some functions to produce partial explanations with TLACEs.
"""

def explain_witness(fsm, state, spec):
    """
    Explain why state of fsm satisfies spec.
    
    state must satisfy spec in fsm. No check is made to ensure that.
    
    Return a partial TLACE node explaining why state of fsm satifies spec.
    This node contains no branch.
    """
    
    if type(spec) is TrueExp:
        # state is its own explanation, no need for annotation
        return PartialTlacenode(state, None, None, None)
        
    elif type(spec) is FalseExp:    
        print("[ERROR] CTLK partial TLACE explain_witness:",
              "cannot explain why state satisfies False",
              spec)
        # TODO Error, cannot explain why state |= False
        return None
        
    elif type(spec) is Init:
        return PartialTlacenode(state, [spec], None, None)
    
    elif type(spec) is Atom:
        # state is its own explanation
        return PartialTlacenode(state, [spec], None, None)
        
    elif type(spec) is Not:
        return explain_countex(fsm, state, spec.child)
        
    elif type(spec) is And:
        # Get left and right explanations, then merge them
        left = explain_witness(fsm, state, spec.left)
        right = explain_witness(fsm, state, spec.right)
        return PartialTlacenode(state,
                         left.atomics + right.atomics,
                         left.branches + right.branches,
                         left.universals + right.universals
                        )
        
    elif type(spec) is Or:
        # If state satisfies spec.left, explain it
        # otherwise, state satisfies spec.right, so explain it
        specbdd = evalCTLK(fsm, spec.left)
        if state <= specbdd:
            return explain_witness(fsm, state, spec.left)
        else:
            return explain_witness(fsm, state, spec.right)
        
    elif type(spec) is Implies:
        # a -> b is ~a | b
        return explain_witness(fsm, state, Or(Not(spec.left), spec.right))
        
    elif type(spec) is Iff:
        # a <-> b is (a & b) | (~a & ~b)
        return explain_witness(
                fsm, state,
                Or(And(spec.left, spec.right),
                   And(Not(spec.left), Not(spec.right)))
               )
        
    elif type(spec) in {AF, AG, AX, AU, AW, K, E, D, C}:
        # Cannot explain with a single path
        return PartialTlacenode(state, None, None, [spec])
                       
    elif type(spec) in {EF, EG, EX, EU, EW, nK, nE, nD, nC, Reachable}:
        # The set of branches MUST be a list (and not a tuple)
        # to be able to explain them later
        return PartialTlacenode(state, None, [spec], None)
        
    else:
        # TODO Generate error
        print("[ERROR] CTLK partial TLACE explain_witness:",
              "unrecognized specification type",
              spec)
        return None
        
        
def explain_countex(fsm, state, spec):
    """
    Explain why state of fsm violates spec.
    
    state must not satisfy spec in fsm. No check is made to ensure that.
    
    Return a partial TLACE node explaining why state of fsm violates spec.
    This node contains no branch.
    """
    
    if type(spec) is TrueExp:    
        print("[ERROR] CTLK partial TLACE explain_countex:",
              "cannot explain why state violates True",
              spec)
        # TODO Error, cannot explain why state |/= True
        return None
        
    elif type(spec) is FalseExp:
        return explain_witness(fsm, state, TrueExp())
        
    elif type(spec) is Init:
        return PartialTlacenode(state, [Not(spec)], None, None)
        
    elif type(spec) is Atom:
        # state is its own explanation
        return PartialTlacenode(state, [Not(spec)], None, None)
        
    elif type(spec) is Not:
        return explain_witness(fsm, state, spec.child)
        
    elif type(spec) is And:
        # ~(a & b) = ~a | ~b
        return explain_witness(fsm, state, Or(Not(spec.left), Not(spec.right)))
        
    elif type(spec) is Or:
        # ~(a | b) = ~a & ~b
        return explain_witness(fsm, state, And(Not(spec.left), Not(spec.right)))
        
    elif type(spec) is Implies:
        # ~(a -> b) = a & ~b
        return explain_witness(fsm, state, And(spec.left, Not(spec.right)))
        
    elif type(spec) is Iff:
        # ~(a <-> b) = (a & ~b) | (~a & b)
        return explain_witness(fsm, state,
                               Or(And(spec.left, Not(spec.right)),
                                  And(Not(spec.left), spec.right)))
        
    elif type(spec) is AF:
        # AF p = ~EG ~p
        return explain_witness(fsm, state, EG(Not(spec.child)))
        
    elif type(spec) is AG:
        # AG p = ~EF ~p
        return explain_witness(fsm, state, EF(Not(spec.child)))
        
    elif type(spec) is AX:
        # AX p = ~EX ~p
        return explain_witness(fsm, state, EX(Not(spec.child)))
        
    elif type(spec) is AU:
        # A[p U q] = ~E[~q W ~p & ~q]
        return explain_witness(fsm, state,
                               EW(Not(spec.right),
                                  And(Not(spec.left), Not(spec.right))))
        
    elif type(spec) is AW:
        # A[p W q] = ~E[~q U ~p & ~q]
        return explain_witness(fsm, state,
                               EU(Not(spec.right),
                                  And(Not(spec.left), Not(spec.right))))
                     
    elif type(spec) is EF:
        # EF p = ~AG ~p
        return explain_witness(fsm, state, AG(Not(spec.child)))
        
    elif type(spec) is EG:
        # EG p = ~AF ~p
        return explain_witness(fsm, state, AF(Not(spec.child)))
        
    elif type(spec) is EX:
        # EX p = ~AX ~p
        return explain_witness(fsm, state, AX(Not(spec.child)))
        
    elif type(spec) is EU:
        # E[p U q] = ~A[~q W ~p & ~q]
        return explain_witness(fsm, state,
                               AW(Not(spec.right),
                                  And(Not(spec.left), Not(spec.right))))
        
    elif type(spec) is EW:
        # E[p W q] = ~A[~q U ~p & ~q]
        return explain_witness(fsm, state,
                               AU(Not(spec.right),
                                  And(Not(spec.left), Not(spec.right))))
                                  
    elif type(spec) is nK:
        # nK<ag> p = ~K<ag> ~p
        return explain_witness(fsm, state, K(spec.agent, Not(spec.child)))
                                  
    elif type(spec) is nE:
        # nE<group> p = ~E<group> ~p
        return explain_witness(fsm, state, E(spec.group, Not(spec.child)))
                                  
    elif type(spec) is nD:
        # nD<group> p = ~D<group> ~p
        return explain_witness(fsm, state, D(spec.group, Not(spec.child)))
                                  
    elif type(spec) is nC:
        # nC<group> p = ~C<group> ~p
        return explain_witness(fsm, state, C(spec.group, Not(spec.child)))
                                  
    elif type(spec) is K:
        # K<ag> p = ~nK<ag> ~p
        return explain_witness(fsm, state, nK(spec.agent, Not(spec.child)))
                                  
    elif type(spec) is E:
        # E<group> p = ~nE<group> ~p
        return explain_witness(fsm, state, nE(spec.group, Not(spec.child)))
                                  
    elif type(spec) is D:
        # D<group> p = ~nD<group> ~p
        return explain_witness(fsm, state, nD(spec.group, Not(spec.child)))
                                  
    elif type(spec) is C:
        # C<group> p = ~nC<group> ~p
        return explain_witness(fsm, state, nC(spec.group, Not(spec.child)))
        
    elif type(spec) is Reachable:
        # Is its own explanation
        return PartialTlacenode(state, None, None, [Not(spec)])
        
    else:
        # TODO Generate error
        print("[ERROR] CTLK partial explain_countex:",
              "unrecognized specification type",
              spec)
        return None
        
        
def explain_branch(fsm, state, spec, originalspec):
    """
    Return a TLACE branch explaining why state of fsm satisfies spec.
    
    state must satisfy spec in fsm. No check is made to ensure that.
    spec must be an existential operator (EX, EF, EU, EG, EW, nK, nE, nD, nC).
    originalspec is the specification with which the created branch will be
    annotated.
    
    The explanation is only about the top operator of spec; all nodes do not
    carry any branch.
    """
    if type(spec) is EX:
        phi = evalCTLK(fsm, spec.child)
        path = explain_ex(fsm, state, phi)
        branch = (PartialTlacenode(path[0]),
                  path[1],
                  explain_witness(fsm, path[2], spec.child))
        return TemporalBranch(originalspec, branch)

    elif type(spec) is EF:
        return explain_branch(fsm, state,
                              EU(TrueExp(), spec.child),
                              originalspec)

    elif type(spec) is EG:
        phi = evalCTLK(fsm, spec.child)
        (path, (inloop, loopstate)) = explain_eg(fsm, state, phi)

        branch = []
        # intermediate states
        loop = None
        for s, i in zip(path[::2], path[1::2]):
            wit = explain_witness(fsm, s, spec.child)
            branch.append(wit)
            branch.append(i)
            # manage the loop
            if loopstate == s:
                loop = wit
                
        # last state and loop
        wit = explain_witness(fsm, path[-1], spec.child)
        branch.append(wit)
        if loopstate == path[-1]:
            loop = wit
        finalloop = (inloop, loop)
        
        return TemporalBranch(originalspec, tuple(branch), finalloop)

    elif type(spec) is EU:
        phi = evalCTLK(fsm, spec.left)
        psi = evalCTLK(fsm, spec.right)
        path = explain_eu(fsm, state, phi, psi)

        branch = []
        # intermediate states
        for s, i in zip(path[::2], path[1::2]):
            branch.append(explain_witness(fsm, s, spec.left))
            branch.append(i)
        # last state
        branch.append(explain_witness(fsm, path[-1], spec.right))

        return TemporalBranch(originalspec, tuple(branch))

    elif type(spec) is EW:
        # E[p W q] = E[p U q] | EG p
        euspec = EU(spec.left, spec.right)
        egspec = EG(spec.left)
        if state <= evalCTLK(fsm, euspec):
            return explain_branch(fsm, state, euspec, originalspec)
        else:
            return explain_branch(fsm, state, egspec, originalspec)
            
    elif type(spec) is Reachable:
        # Get the inversed path
        path = explain_reachable(fsm, state)
        
        # Construct the path of nodes
        branch = []
        for (s, i) in zip(path[::2], path[1::2]):
            branch.append(PartialTlacenode(s))
            branch.append(i)
        # Special case for the last node: it is Init
        branch.append(explain_witness(fsm, path[-1], Init()))
        
        return TemporalBranch(originalspec, tuple(branch))
            
    elif type(spec) is nK:        
        # Get the equivalent state
        phi = evalCTLK(fsm, spec.child)
        path = explain_nk(fsm, state, spec.agent.value, phi)
        
        # Explain why the equiv state satisfies Reachable and phi
        # and construct the branch
        branch = [PartialTlacenode(path[0]), path[1],
                  explain_witness(fsm, path[2], And(Reachable(), spec.child))]
        
        # Return the epistemic branch
        return EpistemicBranch(originalspec, tuple(branch))
        
    elif type(spec) is nE:        
        # Get the equivalent state
        phi = evalCTLK(fsm, spec.child)
        path = explain_ne(fsm, state, [ag.value for ag in spec.group], phi)
        
        # Explain why the equiv state satisfies Reachable and phi
        # and construct the branch
        branch = [PartialTlacenode(path[0]), path[1],
                  explain_witness(fsm, path[2], And(Reachable(), spec.child))]
        
        # Return the epistemic branch
        return EpistemicBranch(originalspec, tuple(branch))
        
    elif type(spec) is nD:        
        # Get the equivalent state
        phi = evalCTLK(fsm, spec.child)
        path = explain_nd(fsm, state, [ag.value for ag in spec.group], phi)
        
        # Explain why the equiv state satisfies Reachable and phi
        # and construct the branch
        branch = [PartialTlacenode(path[0]), path[1],
                  explain_witness(fsm, path[2], And(Reachable(), spec.child))]
        
        # Return the epistemic branch
        return EpistemicBranch(originalspec, tuple(branch))
        
    if type(spec) is nC:
        # Get the knowledge path
        phi = evalCTLK(fsm, spec.child)
        path = explain_nc(fsm, state, [ag.value for ag in spec.group], phi)
        
        # Discard the first element of path: this is state and does not have
        # to be explained reachable
        branch = [PartialTlacenode(path[0]), path[1]]
        path = path[2:]
        
        # Build the branch : show that all intermediate states are reachable
        for (s, ag) in zip(path[::2], path[1::2]):
            branch.append(explain_witness(fsm, s, Reachable()))
            branch.append(ag)
        
        # Complete the branch : show that the last one is reachable
        # satisfies phi
        branch.append(explain_witness(fsm, path[-1],
                                                  And(Reachable(), spec.child)))
        
        # Return the epistemic branch
        return EpistemicBranch(originalspec, tuple(branch))

    else:
        # TODO Generate error
        print("[ERROR] CTLK partial explain_branch:",
              "unrecognized specification type",
              spec)
        return None


import unittest

from pynusmv.init import init_nusmv, deinit_nusmv

from pynusmv_tools.mas import glob

class TestGlob(unittest.TestCase):
    
    def setUp(self):
        init_nusmv()
        
    def tearDown(self):
        glob.reset_globals()
        deinit_nusmv()
        
    def test_dining_crypto(self):
        glob.load_from_file("tests/pynusmv_tools/ctlk/dining-crypto.smv")
        fsm = glob.mas()
        self.assertIsNotNone(fsm)
        
        agents = fsm._epistemic.keys()
        self.assertSetEqual(set(agents), {"c1", "c2", "c3"})


"""
ATLK with partial observability evaluation functions.

These evaluation functions reimplement the algorithms with generators to
decrease the memory consumption of the implementation.
"""

import gc
from functools import reduce

from pynusmv.dd import BDD
from pynusmv.mc import eval_simple_expression
from pynusmv.utils import fixpoint as fp
from pynusmv.exception import PyNuSMVError

from ..mas.mas import Agent, Group

from ..atlkFO.ast import (TrueExp, FalseExp, Init, Reachable,
                          Atom, Not, And, Or, Implies, Iff, 
                          AF, AG, AX, AU, AW, EF, EG, EX, EU, EW,
                          nK, nE, nD, nC, K, E, D, C,
                          CEF, CEG, CEX, CEU, CEW, CAF, CAG, CAX, CAU, CAW)
                          

from ..atlkFO.eval import (fair_states, ex, eg, eu, nk, ne, nd, nc)

from . import config


# A dictionary to keep track of number of strategies for improved algorithm
__strategies = {} 
# A dictionary to keep track of number of filterings for improved algorithm
__filterings = {} 


def evalATLK(fsm, spec, variant="SF", semantics="group"):
    """
    Return the BDD representing the set of states of fsm satisfying spec.
    
    fsm -- a MAS representing the system
    spec -- an AST-based ATLK specification
    variant -- the variant of the algorithm to evaluate strategic operators;
               must be
               * "SF" for the standard way: splitting in uniform strategies then
                 filtering winning states,
               * "FS" for the alternating way: filtering winning states, then
                 splitting one conflicting equivalence class, then recurse
               * "FSF" for the filter-split-filter way: filtering winning states
                 then splitting all remaining actions into uniform strategies,
                 then filtering final winning states.
    semantics -- the semantics to use for starting point and strategy point
                 equivalence; must be
                 * "group" for the original ATLK_irF semantics considering
                   the group as a single agent (distributed knowledge is used)
                 * "individual" for the original ATL_ir semantics considering
                   the group as individual agents (individual knowledge is
                   used)
    
    If variant is not in {"SF", "FS", "FSF"}, the standard "SF" way is used.
    If semantics is not in {"group", "individual"}, the "group" semantics is
    used.
    """
    
    if type(spec) is TrueExp:
        return BDD.true(fsm.bddEnc.DDmanager)
        
    elif type(spec) is FalseExp:
        return BDD.false(fsm.bddEnc.DDmanager)
        
    elif type(spec) is Init:
        return fsm.init
        
    elif type(spec) is Reachable:
        return fsm.reachable_states
    
    elif type(spec) is Atom:
        return eval_simple_expression(fsm, spec.value)
        
    elif type(spec) is Not:
        return ~evalATLK(fsm, spec.child, variant=variant, semantics=semantics)
        
    elif type(spec) is And:
        return (evalATLK(fsm, spec.left, variant=variant, semantics=semantics)
                & evalATLK(fsm, spec.right, variant=variant,
                           semantics=semantics))
        
    elif type(spec) is Or:
        return (evalATLK(fsm, spec.left, variant=variant, semantics=semantics)
                | evalATLK(fsm, spec.right, variant=variant,
                           semantics=semantics))
        
    elif type(spec) is Implies:
        # a -> b = ~a | b
        return ((~evalATLK(fsm, spec.left, variant=variant,
                           semantics=semantics))
                 | evalATLK(fsm, spec.right, variant=variant,
                            semantics=semantics))
        
    elif type(spec) is Iff:
        # a <-> b = (a & b) | (~a & ~b)
        l = evalATLK(fsm, spec.left, variant=variant, semantics=semantics)
        r = evalATLK(fsm, spec.right, variant=variant, semantics=semantics)
        return (l & r) | ((~l) & (~r))
        
    elif type(spec) is EX:
        return ex(fsm, evalATLK(fsm, spec.child, variant=variant,
                                semantics=semantics))
        
    elif type(spec) is AX:
        # AX p = ~EX ~p
        return ~ex(fsm, ~evalATLK(fsm, spec.child, variant=variant,
                                  semantics=semantics))
        
    elif type(spec) is EG:
        return eg(fsm, evalATLK(fsm, spec.child, variant=variant, 
                                semantics=semantics))
        
    elif type(spec) is AG:
        # AG p = ~EF ~p = ~E[ true U ~p ]
        return ~eu(fsm,
                   BDD.true(fsm.bddEnc.DDmanager),
                   ~evalATLK(fsm, spec.child, variant=variant,
                             semantics=semantics))
        
    elif type(spec) is EU:
        return eu(fsm, evalATLK(fsm, spec.left, variant=variant,
                                semantics=semantics),
                       evalATLK(fsm, spec.right, variant=variant, 
                                semantics=semantics))
        
    elif type(spec) is AU:
        # A[p U q] = ~E[~q W ~p & ~q] = ~(E[~q U ~p & ~q] | EG ~q)
        p = evalATLK(fsm, spec.left, variant=variant, semantics=semantics)
        q = evalATLK(fsm, spec.right, variant=variant, semantics=semantics)
        equpq = eu(fsm, ~q, ~q & ~p)
        egq = eg(fsm, ~q)
        return ~(equpq | egq)
        
    elif type(spec) is EF:
        # EF p = E[ true U p ]
        return eu(fsm,
                  BDD.true(fsm.bddEnc.DDmanager),
                  evalATLK(fsm, spec.child, variant=variant,
                           semantics=semantics))
        
    elif type(spec) is AF:
        # AF p = ~EG ~p
        return ~eg(fsm, ~evalATLK(fsm, spec.child, variant=variant, 
                                  semantics=semantics))
        
    elif type(spec) is EW:
        # E[ p W q ] = E[ p U q ] | EG p
        return (eu(fsm, evalATLK(fsm, spec.left, variant=variant,
                                 semantics=semantics),
                        evalATLK(fsm, spec.right, variant=variant,
                                 semantics=semantics)) |
                eg(fsm, evalATLK(fsm, spec.left, variant=variant,
                                 semantics=semantics)))
        
    elif type(spec) is AW:
        # A[p W q] = ~E[~q U ~p & ~q]
        p = evalATLK(fsm, spec.left, variant=variant, semantics=semantics)
        q = evalATLK(fsm, spec.right, variant=variant, semantics=semantics)
        return ~eu(fsm, ~q, ~p & ~q)
        
    elif type(spec) is nK:
        return nk(fsm, spec.agent.value, evalATLK(fsm, spec.child,
                                                  variant=variant,
                                                  semantics=semantics))
        
    elif type(spec) is K:
        # K<'a'> p = ~nK<'a'> ~p
        return ~nk(fsm, spec.agent.value, ~evalATLK(fsm, spec.child,
                                                    variant=variant,
                                                    semantics=semantics))
        
    elif type(spec) is nE:
        return ne(fsm,
                  [a.value for a in spec.group],
                  evalATLK(fsm, spec.child, variant=variant,
                           semantics=semantics))
        
    elif type(spec) is E:
        # E<g> p = ~nE<g> ~p
        return ~ne(fsm,
                   [a.value for a in spec.group],
                   ~evalATLK(fsm, spec.child, variant=variant,
                             semantics=semantics))
        
    elif type(spec) is nD:
        return nd(fsm,
                  [a.value for a in spec.group],
                  evalATLK(fsm, spec.child, variant=variant,
                           semantics=semantics))
        
    elif type(spec) is D:
        # D<g> p = ~nD<g> ~p
        return ~nd(fsm,
                   [a.value for a in spec.group],
                   ~evalATLK(fsm, spec.child, variant=variant,
                             semantics=semantics))
        
    elif type(spec) is nC:
        return nc(fsm,
                  [a.value for a in spec.group],
                  evalATLK(fsm, spec.child, variant=variant,
                           semantics=semantics))
        
    elif type(spec) is C:
        # C<g> p = ~nC<g> ~p
        return ~nc(fsm,
                   [a.value for a in spec.group],
                   ~evalATLK(fsm, spec.child, variant=variant,
                             semantics=semantics))
    
    elif type(spec) is CAX:
        # [g] X p = ~<g> X ~p
        newspec = CEX(spec.group, Not(spec.child))
        return ~evalATLK(fsm, newspec, variant=variant, semantics=semantics)
        
    elif type(spec) is CAG:
        # [g] G p = ~<g> F ~p
        newspec = CEF(spec.group, Not(spec.child))
        return ~evalATLK(fsm, newspec, variant=variant, semantics=semantics)
        
    elif type(spec) is CAU:
        # [g][p U q] = ~<g>[ ~q W ~p & ~q ]
        newspec = CEW(spec.group,
                      Not(spec.right),
                      And(Not(spec.left), Not(spec.right)))
        return ~evalATLK(fsm, newspec, variant=variant, semantics=semantics)
        
    elif type(spec) is CAF:
        # [g] F p = ~<g> G ~p
        newspec = CEG(spec.group, Not(spec.child))
        return ~evalATLK(fsm, newspec, variant=variant, semantics=semantics)
        
    elif type(spec) is CAW:
        # [g][p W q] = ~<g>[~q U ~p & ~q]
        newspec = CEU(spec.group,
                      Not(spec.right),
                      And(Not(spec.left), Not(spec.right)))
        return ~evalATLK(fsm, newspec, variant=variant, semantics=semantics)
                   
    elif type(spec) in {CEX, CEG, CEU, CEF, CEW}:
        if variant == "SF":
            return eval_strat(fsm, spec, semantics=semantics)
        elif variant == "FS":
            raise PyNuSMVError("Eval Gen SI does not support FS variant")
        elif variant == "FSF":
            return eval_strat_FSF(fsm, spec, semantics=semantics)
        else:
            return eval_strat(fsm, spec, semantics=semantics)
        
    else:
        # TODO Generate error
        print("[ERROR] evalATLK: unrecognized specification type", spec)
        return None



def cex_si(fsm, agents, phi, strat=None):
    """
    Return the set of state/inputs pairs of strat satisfying <agents> X phi
    under full observability in strat.
    If strat is None, strat is considered true.
    
    fsm -- a MAS representing the system
    agents -- a list of agents names
    phi -- a BDD representing the set of states of fsm satisfying phi
    strat -- a BDD representing allowed state/inputs pairs, or None
    """
    if not strat:
        strat = BDD.true(fsm.bddEnc.DDmanager)
    
    phi = phi & fsm.bddEnc.statesInputsMask
    
    return fsm.pre_strat_si(phi | nfair_gamma_si(fsm, agents, strat),
                            agents, strat)
    

def ceu_si(fsm, agents, phi, psi, strat=None):
    """
    Return the set of state/inputs pairs of strat satisfying <agents>[phi U psi]
    under full observability in strat.
    If strat is None, strat is considered true.
    
    fsm -- a MAS representing the system
    agents -- a list of agents names
    phi -- a BDD representing the set of states of fsm satisfying phi
    psi -- a BDD representing the set of states of fsm satisfying psi
    strat -- a BDD representing allowed state/inputs pairs, or None
    
    """
    if not strat:
        strat = BDD.true(fsm.bddEnc.DDmanager)
    
    phi = phi & fsm.bddEnc.statesInputsMask & strat
    psi = psi & fsm.bddEnc.statesInputsMask & strat
    
    if len(fsm.fairness_constraints) == 0:
        return fp(lambda Z : psi | (phi & fsm.pre_strat_si(Z, agents, strat)),
                  BDD.false(fsm.bddEnc.DDmanager))
    else:
        nfair = nfair_gamma_si(fsm, agents, strat)
        def inner(Z):
            res = psi
            for f in fsm.fairness_constraints:
                nf = ~f & fsm.bddEnc.statesMask & strat
                res = res | fsm.pre_strat_si(fp(lambda Y :
                                                 (phi | psi | nfair) &
                                                 (Z | nf) &
                                                 (psi |
                                                  fsm.pre_strat_si(Y, agents,
                                                                   strat)
                                                 ),
                                              BDD.true(fsm.bddEnc.DDmanager)),
                                              agents, strat)
            return (psi | phi | nfair) & res
        return fp(inner, BDD.false(fsm.bddEnc.DDmanager))
    

def cew_si(fsm, agents, phi, psi, strat=None):
    """
    Return the set of state/inputs pairs of strat satisfying <agents>[phi W psi]
    under full observability in strat.
    If strat is None, strat is considered true.
    
    fsm -- a MAS representing the system
    agents -- a list of agents names
    phi -- a BDD representing the set of states of fsm satisfying phi
    psi -- a BDD representing the set of states of fsm satisfying psi
    strat -- a BDD representing allowed state/inputs pairs, or None
    
    """
    if not strat:
        strat = BDD.true(fsm.bddEnc.DDmanager)
    
    phi = phi & fsm.bddEnc.statesInputsMask & strat
    psi = psi & fsm.bddEnc.statesInputsMask & strat
    
    nfair = nfair_gamma_si(fsm, agents, strat)
    
    return fp(lambda Y : (psi | phi | nfair) &
                         (psi | fsm.pre_strat_si(Y, agents, strat)),
              BDD.true(fsm.bddEnc.DDmanager))
    
    
def ceg_si(fsm, agents, phi, strat=None):
    """
    Return the set of state/inputs pairs of strat satisfying <agents> G phi
    under full observability in strat.
    If strat is None, strat is considered true.
    
    fsm -- a MAS representing the system
    agents -- a list of agents names
    phi -- a BDD representing the set of states of fsm satisfying phi
    strat -- a BDD representing allowed state/inputs pairs, or None
    
    """
    if not strat:
        strat = BDD.true(fsm.bddEnc.DDmanager)
    
    phi = phi & fsm.bddEnc.statesInputsMask & strat
    
    nfair = nfair_gamma_si(fsm, agents, strat)
    
    return fp(lambda Y : (phi | nfair) & fsm.pre_strat_si(Y, agents, strat),
              BDD.true(fsm.bddEnc.DDmanager))


def nfair_gamma_si(fsm, agents, strat=None):
    """
    Return the set of state/inputs pairs of strat
    in which agents can avoid a fair path in strat.
    If strat is None, it is considered true.
    
    fsm -- the model
    agents -- a list of agents names
    strat -- a BDD representing allowed state/inputs pairs, or None
    
    """
    if not strat:
        strat = BDD.true(fsm.bddEnc.DDmanager)
    
    if len(fsm.fairness_constraints) == 0:
        return BDD.false(fsm.bddEnc.DDmanager)
    else:
        def inner(Z):
            res = BDD.false(fsm.bddEnc.DDmanager)
            for f in fsm.fairness_constraints:
                nf = ~f & fsm.bddEnc.statesMask & strat
                res = res | fsm.pre_strat_si(fp(lambda Y :
                                                 (Z | nf) &
                                                 fsm.pre_strat_si(Y, agents,
                                                                  strat),
                                             BDD.true(fsm.bddEnc.DDmanager)),
                                             agents, strat)
            return res
        return fp(inner, BDD.false(fsm.bddEnc.DDmanager))


def get_equiv_class(fsm, gamma, state):
    """
    Return the equivalence class for gamma state belongs to. The distributed
    knowledge based equivalence fo gamma containing state is returned.
    
    fsm -- the model
    gamma -- a set of agents of fsm
    """
    return fsm.equivalent_states(state, gamma) & fsm.reachable_states


def is_conflicting(fsm, eqclass, gamma):
    """
    Return whether the given equivalence class for gamma is conflicting or not.
    
    fsm -- the model
    eqclass -- a BDD representing a set of states/inputs pairs
    gamma -- a set of agents of fsm
    
    eqclass must be a distributed knowledge based equivalence class.
    
    """
    si = fsm.pick_one_state_inputs(eqclass)
    ngamma_cube = fsm.bddEnc.inputsCube - fsm.inputs_cube_for_agents(gamma)
    return (eqclass - (eqclass & si.forsome(fsm.bddEnc.statesCube |
                                            ngamma_cube))).isnot_false()


def split_conflicting(fsm, eqclass, gamma):
    """
    Split the given equivalence class for gamma into non conflicting
    sub-strategies.
    
    fsm -- the model
    eqclass -- a BDD representing a set of states/inputs pairs
    gamma -- a set of agents of fsm
    
    eqclass must be a distributed knowledge based equivalence class.
    
    Generate all splitted non conflicting sub-strategies.
    
    """
    ngamma_cube = fsm.bddEnc.inputsCube - fsm.inputs_cube_for_agents(gamma)
    # Split eqclass into non-conflicting subsets
    while eqclass.isnot_false():
        si = fsm.pick_one_state_inputs(eqclass)
        ncss = eqclass & si.forsome(fsm.bddEnc.statesCube | ngamma_cube)
        eqclass = eqclass - ncss
    
        yield ncss


def split_one(fsm, strats, gamma):
    """
    Split one equivalence class of strats and return triples composed of
    the common non-conflicting part already encountered,
    a split and the rest of strats to split.
    
    fsm -- the model
    strats -- a BDD representing a set of states/inputs pairs
    gamma -- a set of agents of fsm
    
    Return a generator of all triples of common parts, splits
    and rest of strats.
    
    """
    if strats.is_false():
        yield (strats, strats, strats)
        return
        
    else:
        common = BDD.false(fsm.bddEnc.DDmanager)
        
        while strats.isnot_false():
            # Get one equivalence class
            si = fsm.pick_one_state_inputs(strats)
            s = si.forsome(fsm.bddEnc.inputsCube)
            eqs = get_equiv_class(fsm, gamma, s)
            eqcl = strats & eqs
            
            # Remove it from strats
            strats = strats - eqcl
            
            # The current equivalence class is conflicting
            if is_conflicting(fsm, eqcl, gamma):
                for non_conflicting in split_conflicting(fsm, eqcl, gamma):
                    yield (common, non_conflicting, strats)
                return
            
            else:
                # Add equivalence class to common
                common = common | eqcl
        
        # strats is false, everything is in common
        yield (common, strats, strats)


def split(fsm, strats, gamma, semantics="group"):
    """
    Split strats into all its non-conflicting greatest subsets.
    
    fsm -- the model
    strats -- a BDD representing a set of states/inputs pairs
    gamma -- a set of agents of fsm
    semantics -- the semantic to use for splitting (group or individual)
    
    Return a generator of all non-conflicting greatest subsets of strats.
    
    """
    if semantics == "individual":
        if len(gamma) <= 0:
            yield strats
        else:
            agent = next(iter(gamma))
            gamma = gamma - {agent}
            for strat in split(fsm, strats, {agent}, semantics="group"):
                for sstrat in split(fsm, strat, gamma,
                                    semantics="individual"):
                    yield sstrat
    else:
        if strats.is_false():
            yield strats
        else:
            # semantics should be group
            # Split one equivalence class
            for common, splitted, rest in split_one(fsm, strats, gamma):
                for strat in split(fsm, rest, gamma):
                    yield (common | strat | splitted)


def filter_strat(fsm, spec, strat=None, variant="SF", semantics="group"):
    """
    Returns the subset SA of strat (or the whole system if strat is None),
    state/action pairs of fsm, such that there is a strategy to satisfy spec
    in fsm.
    
    fsm -- a MAS representing the system;
    spec -- an AST-based ATLK specification with a top strategic operator;
            the operator is CEX, CEG, CEF, CEU or CEW.
    strat -- the subset of the system to consider.
    variant -- the variant of the algorithm to evaluate strategic operators;
               must be
               * "SF" for the standard way: splitting in uniform strategies then
                 filtering winning states,
               * "FS" for the alternating way: filtering winning states, then
                 splitting one conflicting equivalence class, then recurse
               * "FSF" for the filter-split-filter way: filtering winning states
                 then splitting all remaining actions into uniform strategies,
                 then filtering final winning states.
    semantics -- the semantics to use for starting point and strategy point
                 equivalence; must be
                 * "group" for the original ATLK_irF semantics considering
                   the group as a single agent (distributed knowledge is used)
                 * "individual" for the original ATL_ir semantics considering
                   the group as individual agents (individual knowledge is
                   used)
                 
    If variant is not in {"SF", "FS", "FSF"}, the standard "SF" way is used.
    """
    
    sat = BDD.false(fsm.bddEnc.DDmanager)
    agents = {atom.value for atom in spec.group}
    
    # Filtering
    if type(spec) is CEX:
        winning = cex_si(fsm, agents,
                         evalATLK(fsm, spec.child, variant=variant,
                                  semantics=semantics), strat)

    elif type(spec) is CEG:
        winning = ceg_si(fsm, agents, evalATLK(fsm, spec.child,
                                               variant=variant,
                                               semantics=semantics), strat)

    elif type(spec) is CEU:
        winning = ceu_si(fsm, agents, evalATLK(fsm, spec.left,
                                               variant=variant,
                                               semantics=semantics),
                         evalATLK(fsm, spec.right, variant=variant,
                                  semantics=semantics), strat)

    elif type(spec) is CEF:
        # <g> F p = <g>[true U p]
        winning = ceu_si(fsm, agents, BDD.true(fsm.bddEnc.DDmanager),
                         evalATLK(fsm, spec.child, variant=variant,
                                  semantics=semantics), strat)

    elif type(spec) is CEW:
       winning = cew_si(fsm, agents, evalATLK(fsm, spec.left, variant=variant,
                                              semantics=semantics),
                        evalATLK(fsm, spec.right, variant=variant,
                                 semantics=semantics), strat)
    
    
    return winning & fsm.bddEnc.statesInputsMask & fsm.protocol(agents)


def all_equiv_sat(fsm, winning, agents, semantics="group"):
    """
    Return the states s of winning such that all states indistinguishable
    from s by agents are in winning.
    
    fsm -- a MAS representing the system;
    winning -- a BDD representing a set of states of fsm;
    agents -- a set of agents;
    semantics -- the semantics to use (group or individual); if not
                 "individual", "group" semantics is used.
    
    """
    if semantics == "individual":
        equiv_sat = BDD.true(fsm.bddEnc.DDmanager)
        for agent in agents:
            equiv_sat &= all_equiv_sat(fsm, winning, {agent},
                                       semantics="group")
        return equiv_sat
    else:
        # Get states for which all states belong to winning
        # wineq is the set of states for which all equiv states are in winning
        nwinning = ~winning & fsm.bddEnc.statesInputsMask
        return ~(fsm.equivalent_states(nwinning &
                 fsm.reachable_states, frozenset(agents))) & winning


def agents_in_group(fsm, group):
    """
    Return the set of agents in the given group in fsm. If group is not a group
    of fsm, group is returned alone.
    
    This procedure recursively searches for all basic agents in the groups
    possibly composing group.
    
    fsm -- the model;
    group -- a group or agent name of the model.
    """
    
    if group in fsm.groups:
        agents = set()
        for agent in fsm.groups[group]:
            if agent in fsm.groups:
                agents |= agents_in_group(fsm, agent)
            else:
                agents.add(agent)
    else:
        agents = {group}
    return agents


def eval_strat(fsm, spec, semantics="group"):
    """
    Return the BDD representing the set of states of fsm satisfying spec.
    spec is a strategic operator <G> pi.
    
    fsm -- a MAS representing the system;
    spec -- an AST-based ATLK specification with a top strategic operator.
    semantics -- the semantics to use (group or individual)
    
    """
    sat = BDD.false(fsm.bddEnc.DDmanager)
    agents = {atom.value for atom in spec.group}
    
    if semantics == "individual":
        agents = reduce(lambda a, b: a | b,
                        (agents_in_group(fsm, group) for group in agents))
    
    # Restrict protocol to reachable states to avoid splitting useless
    # equivalence classes
    protocol = fsm.protocol(agents) & fsm.reachable_states
    strats = split(fsm, protocol, agents, semantics=semantics)
    nbstrats = 0
    
    if config.debug:
        moves_count = fsm.count_states_inputs(protocol)
        print("Eval strategies (SF): {} move{} in protocol"
              .format(moves_count, "s" if moves_count > 1 else ""))
    
    for strat in strats:
        nbstrats += 1
        winning = (filter_strat(fsm, spec, strat, variant="SF").
                    forsome(fsm.bddEnc.inputsCube))
        sat = sat | all_equiv_sat(fsm, winning, agents, semantics=semantics)
        
        # ----- Garbage collection -------------------------------------
        if (config.garbage.type == "each" or
            (config.garbage.type == "step"
                and nbstrats % config.garbage.step == 0)):
            gc.collect()
        
        if config.debug and nbstrats % 1000 == 0:
            print("Eval strategies (SF): {} strateg{} checked so far"
                  .format(nbstrats, "ies" if nbstrats > 1 else "y"))
    
    # DEBUG Print number of strategies
    if config.debug:
        print("Eval_strat: {} strategies".format(nbstrats))
    
    return sat
    
    
def eval_strat_FSF(fsm, spec, semantics="group"):
    """
    Return the BDD representing the set of states of fsm satisfying spec.
    spec is a strategic operator <G> pi.
    
    Implement a variant of the algorithm that
    filters, splits and then filters.
    
    fsm -- a MAS representing the system;
    spec -- an AST-based ATLK specification with a top strategic operator.
    semantics -- the semantics to use (group or individual).
    
    """
    
    sat = BDD.false(fsm.bddEnc.DDmanager)
    agents = {atom.value for atom in spec.group}
    
    if semantics == "individual":
        agents = reduce(lambda a, b: a | b,
                        (agents_in_group(fsm, group) for group in agents))
    
    protocol = fsm.protocol(agents) & fsm.reachable_states
    
    if config.debug:
        moves_count = fsm.count_states_inputs(protocol)
        print("Eval strategies (FSF): {} move{} in protocol"
              .format(moves_count, "s" if moves_count > 1 else ""))
    
    # First filtering
    winning = filter_strat(fsm, spec, strat=protocol, variant="FSF",
                           semantics=semantics)
    
    if config.debug:
        moves_count = fsm.count_states_inputs(winning)
        print("Eval strategies (FSF): {} move{} after filtering"
              .format(moves_count, "s" if moves_count > 1 else ""))
    
    if winning.is_false(): # no state/inputs pairs are winning => return false
        return winning
    
    # Splitting the strategies
    nbstrats = 0
    for strat in split(fsm, winning, agents, semantics=semantics):
        nbstrats += 1
        
        if config.debug and nbstrats % 1000 == 0:
            print("Eval strategies (FSF): {} strateg{} checked so far"
                  .format(nbstrats, "ies" if nbstrats > 1 else "y"))
        
        # Second filtering
        winning = filter_strat(fsm, spec, strat, variant="FSF",
                               semantics=semantics)
        winning = winning.forsome(fsm.bddEnc.inputsCube)
        sat = sat | all_equiv_sat(fsm, winning, agents, semantics=semantics)
        
        # Collect to avoid memory overflow
        if (config.garbage.type == "each" or
            (config.garbage.type == "step"
             and nbstrats % config.garbage.step == 0)):
            gc.collect()
    
    if config.debug:
        print("Eval strategies (FSF): {} strateg{} checked"
              .format(nbstrats, "ies" if nbstrats > 1 else "y"))
    
    return sat


import unittest

from pynusmv.dd import BDD
from pynusmv.init import init_nusmv, deinit_nusmv
from pynusmv.mc import eval_simple_expression

from pynusmv_tools.mas import glob
from pynusmv_tools.atlkFO.eval import (evalATLK, cax, cag, cau, caw, fair_gamma_states)


class TestEval(unittest.TestCase):
    
    def setUp(self):
        init_nusmv()
    
    def tearDown(self):
        glob.reset_globals()
        deinit_nusmv()
    
    
    def transmission(self):
        glob.load_from_file("tests/pynusmv_tools/atlkFO/models/transmission.smv")
        fsm = glob.mas()
        self.assertIsNotNone(fsm)
        return fsm
        
    
    def transmission_fair(self):
        glob.load_from_file("tests/pynusmv_tools/atlkFO/models/transmission-fair.smv")
        fsm = glob.mas()
        self.assertIsNotNone(fsm)
        return fsm
        
    
    def cardgame(self):
        glob.load_from_file("tests/pynusmv_tools/atlkFO/models/cardgame.smv")
        fsm = glob.mas()
        self.assertIsNotNone(fsm)
        return fsm
        
    
    def cardgame_fair(self):
        glob.load_from_file("tests/pynusmv_tools/atlkFO/models/cardgame-fair.smv")
        fsm = glob.mas()
        self.assertIsNotNone(fsm)
        return fsm
        
        
    def test_cax(self):
        fsm = self.cardgame()
        
        s0 = eval_simple_expression(fsm, "step = 0")
        s1 = eval_simple_expression(fsm, "step = 1")
        s2 = eval_simple_expression(fsm, "step = 2")
        pa = eval_simple_expression(fsm, "pcard = Ac")
        pk = eval_simple_expression(fsm, "pcard = K")
        pq = eval_simple_expression(fsm, "pcard = Q")
        da = eval_simple_expression(fsm, "dcard = Ac")
        dk = eval_simple_expression(fsm, "dcard = K")
        dq = eval_simple_expression(fsm, "dcard = Q")
        dda = eval_simple_expression(fsm, "ddcard = Ac")
        ddk = eval_simple_expression(fsm, "ddcard = K")
        ddq = eval_simple_expression(fsm, "ddcard = Q")
        win = eval_simple_expression(fsm, "win")
        lose = eval_simple_expression(fsm, "lose")
        true = eval_simple_expression(fsm, "TRUE")
        false = eval_simple_expression(fsm, "FALSE")
        
        self.assertTrue(s0 & fsm.reachable_states <= cax(fsm, {"player"}, s1))
        self.assertTrue(s1 & fsm.reachable_states <= cax(fsm, {"dealer"}, s2))
        
        self.assertFalse(s1 & pk & fsm.reachable_states <= cax(fsm, {"player"},
                                                               s2 & pk))
        self.assertTrue(s1 & dk <= cax(fsm, {"dealer"}, s2 & dk))